"""Module for HttpClientSession class."""

from __future__ import annotations

import asyncio
import logging
import ssl
import time
from typing import Any

import aiohttp
from yarl import URL

from .deviceconfig import DeviceConfig
from .exceptions import (
    KasaException,
    TimeoutError,
    _ConnectionError,
)
from .json import loads as json_loads

_LOGGER = logging.getLogger(__name__)


def get_cookie_jar() -> aiohttp.CookieJar:
    """Return a new cookie jar with the correct options for device communication."""
    return aiohttp.CookieJar(unsafe=True, quote_cookie=False)


class HttpClient:
    """HttpClient Class."""

    # Some devices (only P100 so far) close the http connection after each request
    # and aiohttp doesn't seem to handle it. If a Client OS error is received the
    # http client will start ensuring that sequential requests have a wait delay.
    WAIT_BETWEEN_REQUESTS_ON_OSERROR = 0.25

    def __init__(self, config: DeviceConfig) -> None:
        self._config = config
        self._client_session: aiohttp.ClientSession | None = None
        self._jar = aiohttp.CookieJar(unsafe=True, quote_cookie=False)
        self._last_url = URL(f"http://{self._config.host}/")

        self._wait_between_requests = 0.0
        self._last_request_time = 0.0

    @property
    def client(self) -> aiohttp.ClientSession:
        """Return the underlying http client."""
        if self._config.http_client and issubclass(
            self._config.http_client.__class__, aiohttp.ClientSession
        ):
            return self._config.http_client

        if not self._client_session:
            self._client_session = aiohttp.ClientSession(cookie_jar=get_cookie_jar())
        return self._client_session

    async def post(
        self,
        url: URL,
        *,
        params: dict[str, Any] | None = None,
        data: bytes | None = None,
        json: dict | Any | None = None,
        headers: dict[str, str] | None = None,
        cookies_dict: dict[str, str] | None = None,
        ssl: ssl.SSLContext | bool = False,
    ) -> tuple[int, dict | bytes | None]:
        """Send an http post request to the device.

        If the request is provided via the json parameter json will be returned.
        """
        # Once we know a device needs a wait between sequential queries always wait
        # first rather than keep erroring then waiting.
        if self._wait_between_requests:
            now = time.monotonic()
            gap = now - self._last_request_time
            if gap < self._wait_between_requests:
                sleep = self._wait_between_requests - gap
                _LOGGER.debug(
                    "Device %s waiting %s seconds to send request",
                    self._config.host,
                    sleep,
                )
                await asyncio.sleep(sleep)

        _LOGGER.debug("Posting to %s", url)
        response_data = None
        self._last_url = url
        self.client.cookie_jar.clear()
        return_json = bool(json)
        if self._config.timeout is None:
            _LOGGER.warning("Request timeout is set to None.")
        client_timeout = aiohttp.ClientTimeout(total=self._config.timeout)

        # If json is not a dict send as data.
        # This allows the json parameter to be used to pass other
        # types of data such as async_generator and still have json
        # returned.
        if json and not isinstance(json, dict):
            data = json
            json = None
        try:
            resp = await self.client.post(
                url,
                params=params,
                data=data,
                json=json,
                timeout=client_timeout,
                cookies=cookies_dict,
                headers=headers,
                ssl=ssl,
            )
            async with resp:
                response_data = await resp.read()

            if resp.status == 200:
                if return_json:
                    response_data = json_loads(response_data.decode())
            else:
                _LOGGER.debug(
                    "Device %s received status code %s with response %s",
                    self._config.host,
                    resp.status,
                    str(response_data),
                )
                if response_data and return_json:
                    try:
                        response_data = json_loads(response_data.decode())
                    except Exception:
                        _LOGGER.debug("Device %s response could not be parsed as json")

        except (aiohttp.ServerDisconnectedError, aiohttp.ClientOSError) as ex:
            if not self._wait_between_requests:
                _LOGGER.debug(
                    "Device %s received an os error, "
                    "enabling sequential request delay: %s",
                    self._config.host,
                    ex,
                )
                self._wait_between_requests = self.WAIT_BETWEEN_REQUESTS_ON_OSERROR
            self._last_request_time = time.monotonic()
            raise _ConnectionError(
                f"Device connection error: {self._config.host}: {ex}", ex
            ) from ex
        except (aiohttp.ServerTimeoutError, TimeoutError) as ex:
            raise TimeoutError(
                "Unable to query the device, "
                + f"timed out: {self._config.host}: {ex}",
                ex,
            ) from ex
        except Exception as ex:
            raise KasaException(
                f"Unable to query the device: {self._config.host}: {ex}", ex
            ) from ex

        # For performance only request system time if waiting is enabled
        if self._wait_between_requests:
            self._last_request_time = time.monotonic()

        return resp.status, response_data

    def get_cookie(self, cookie_name: str) -> str | None:
        """Return the cookie with cookie_name."""
        if cookie := self.client.cookie_jar.filter_cookies(self._last_url).get(
            cookie_name
        ):
            return cookie.value
        return None

    async def close(self) -> None:
        """Close the ClientSession."""
        client = self._client_session
        self._client_session = None
        if client:
            await client.close()



import pytest
from pytest_mock import MockerFixture

from kasa import DeviceType, Module
from kasa.cli.hub import hub

from ..device_fixtures import hubs, plug_iot


@hubs
async def test_hub_pair(dev, mocker: MockerFixture, runner, caplog):
    """Test that pair calls the expected methods."""
    cs = dev.modules.get(Module.ChildSetup)
    # Patch if the device supports the module
    if cs is not None:
        mock_pair = mocker.patch.object(cs, "pair")

    res = await runner.invoke(hub, ["pair"], obj=dev, catch_exceptions=False)
    if cs is None:
        assert "is not a hub" in res.output
        return

    mock_pair.assert_awaited()
    assert "Finding new devices for 10 seconds" in res.output
    assert res.exit_code == 0


@hubs
async def test_hub_unpair(dev, mocker: MockerFixture, runner):
    """Test that unpair calls the expected method."""
    if not dev.children:
        pytest.skip("Cannot test without child devices")

    id_ = next(iter(dev.children)).device_id

    cs = dev.modules.get(Module.ChildSetup)
    mock_unpair = mocker.spy(cs, "unpair")

    res = await runner.invoke(hub, ["unpair", id_], obj=dev, catch_exceptions=False)

    mock_unpair.assert_awaited()
    assert f"Unpaired {id_}" in res.output
    assert res.exit_code == 0


@plug_iot
async def test_non_hub(dev, mocker: MockerFixture, runner):
    """Test that hub commands return an error if executed on a non-hub."""
    assert dev.device_type is not DeviceType.Hub
    res = await runner.invoke(
        hub, ["unpair", "dummy_id"], obj=dev, catch_exceptions=False
    )
    assert "is not a hub" in res.output



"""Module for base energy module."""

from __future__ import annotations

from abc import ABC, abstractmethod
from enum import IntFlag, auto
from typing import TYPE_CHECKING, Any
from warnings import warn

from ..emeterstatus import EmeterStatus
from ..feature import Feature
from ..module import Module


class Energy(Module, ABC):
    """Base interface to represent an Energy module."""

    class ModuleFeature(IntFlag):
        """Features supported by the device."""

        #: Device reports :attr:`voltage` and :attr:`current`
        VOLTAGE_CURRENT = auto()
        #: Device reports :attr:`consumption_total`
        CONSUMPTION_TOTAL = auto()
        #: Device reports periodic stats via :meth:`get_daily_stats`
        #: and :meth:`get_monthly_stats`
        PERIODIC_STATS = auto()

    _supported: ModuleFeature = ModuleFeature(0)

    def supports(self, module_feature: Energy.ModuleFeature) -> bool:
        """Return True if module supports the feature."""
        return module_feature in self._supported

    def _initialize_features(self) -> None:
        """Initialize features."""
        device = self._device
        self._add_feature(
            Feature(
                device,
                name="Current consumption",
                attribute_getter="current_consumption",
                container=self,
                unit_getter=lambda: "W",
                id="current_consumption",
                precision_hint=1,
                category=Feature.Category.Primary,
                type=Feature.Type.Sensor,
            )
        )
        self._add_feature(
            Feature(
                device,
                name="Today's consumption",
                attribute_getter="consumption_today",
                container=self,
                unit_getter=lambda: "kWh",
                id="consumption_today",
                precision_hint=3,
                category=Feature.Category.Info,
                type=Feature.Type.Sensor,
            )
        )
        self._add_feature(
            Feature(
                device,
                id="consumption_this_month",
                name="This month's consumption",
                attribute_getter="consumption_this_month",
                container=self,
                unit_getter=lambda: "kWh",
                precision_hint=3,
                category=Feature.Category.Info,
                type=Feature.Type.Sensor,
            )
        )
        if self.supports(self.ModuleFeature.CONSUMPTION_TOTAL):
            self._add_feature(
                Feature(
                    device,
                    name="Total consumption since reboot",
                    attribute_getter="consumption_total",
                    container=self,
                    unit_getter=lambda: "kWh",
                    id="consumption_total",
                    precision_hint=3,
                    category=Feature.Category.Info,
                    type=Feature.Type.Sensor,
                )
            )
        if self.supports(self.ModuleFeature.VOLTAGE_CURRENT):
            self._add_feature(
                Feature(
                    device,
                    name="Voltage",
                    attribute_getter="voltage",
                    container=self,
                    unit_getter=lambda: "V",
                    id="voltage",
                    precision_hint=1,
                    category=Feature.Category.Primary,
                    type=Feature.Type.Sensor,
                )
            )
            self._add_feature(
                Feature(
                    device,
                    name="Current",
                    attribute_getter="current",
                    container=self,
                    unit_getter=lambda: "A",
                    id="current",
                    precision_hint=2,
                    category=Feature.Category.Primary,
                    type=Feature.Type.Sensor,
                )
            )

    @property
    @abstractmethod
    def status(self) -> EmeterStatus:
        """Return current energy readings."""

    @property
    @abstractmethod
    def current_consumption(self) -> float | None:
        """Get the current power consumption in Watt."""

    @property
    @abstractmethod
    def consumption_today(self) -> float | None:
        """Return today's energy consumption in kWh."""

    @property
    @abstractmethod
    def consumption_this_month(self) -> float | None:
        """Return this month's energy consumption in kWh."""

    @property
    @abstractmethod
    def consumption_total(self) -> float | None:
        """Return total consumption since last reboot in kWh."""

    @property
    @abstractmethod
    def current(self) -> float | None:
        """Return the current in A."""

    @property
    @abstractmethod
    def voltage(self) -> float | None:
        """Get the current voltage in V."""

    @abstractmethod
    async def get_status(self) -> EmeterStatus:
        """Return real-time statistics."""

    @abstractmethod
    async def erase_stats(self) -> dict:
        """Erase all stats."""

    @abstractmethod
    async def get_daily_stats(
        self, *, year: int | None = None, month: int | None = None, kwh: bool = True
    ) -> dict:
        """Return daily stats for the given year & month.

        The return value is a dictionary of {day: energy, ...}.
        """

    @abstractmethod
    async def get_monthly_stats(
        self, *, year: int | None = None, kwh: bool = True
    ) -> dict:
        """Return monthly stats for the given year."""

    _deprecated_attributes = {
        "emeter_today": "consumption_today",
        "emeter_this_month": "consumption_this_month",
        "realtime": "status",
        "get_realtime": "get_status",
        "erase_emeter_stats": "erase_stats",
        "get_daystat": "get_daily_stats",
        "get_monthstat": "get_monthly_stats",
    }

    if not TYPE_CHECKING:

        def __getattr__(self, name: str) -> Any:
            if attr := self._deprecated_attributes.get(name):
                msg = f"{name} is deprecated, use {attr} instead"
                warn(msg, DeprecationWarning, stacklevel=2)
                return getattr(self, attr)
            raise AttributeError(f"Energy module has no attribute {name!r}")



"""Module for light strip effects (LB*, KL*, KB*)."""

from __future__ import annotations

from typing import cast

EFFECT_AURORA = {
    "custom": 0,
    "id": "xqUxDhbAhNLqulcuRMyPBmVGyTOyEMEu",
    "brightness": 100,
    "name": "Aurora",
    "segments": [0],
    "expansion_strategy": 1,
    "enable": 1,
    "type": "sequence",
    "duration": 0,
    "transition": 1500,
    "direction": 4,
    "spread": 7,
    "repeat_times": 0,
    "sequence": [[120, 100, 100], [240, 100, 100], [260, 100, 100], [280, 100, 100]],
}
EFFECT_BUBBLING_CAULDRON = {
    "custom": 0,
    "id": "tIwTRQBqJpeNKbrtBMFCgkdPTbAQGfRP",
    "brightness": 100,
    "name": "Bubbling Cauldron",
    "segments": [0],
    "expansion_strategy": 1,
    "enable": 1,
    "type": "random",
    "hue_range": [100, 270],
    "saturation_range": [80, 100],
    "brightness_range": [50, 100],
    "duration": 0,
    "transition": 200,
    "init_states": [[270, 100, 100]],
    "fadeoff": 1000,
    "random_seed": 24,
    "backgrounds": [[270, 40, 50]],
}
EFFECT_CANDY_CANE = {
    "custom": 0,
    "id": "HCOttllMkNffeHjEOLEgrFJjbzQHoxEJ",
    "brightness": 100,
    "name": "Candy Cane",
    "segments": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15],
    "expansion_strategy": 1,
    "enable": 1,
    "type": "sequence",
    "duration": 700,
    "transition": 500,
    "direction": 1,
    "spread": 1,
    "repeat_times": 0,
    "sequence": [
        [0, 0, 100],
        [0, 0, 100],
        [360, 81, 100],
        [0, 0, 100],
        [0, 0, 100],
        [360, 81, 100],
        [360, 81, 100],
        [0, 0, 100],
        [0, 0, 100],
        [360, 81, 100],
        [360, 81, 100],
        [360, 81, 100],
        [360, 81, 100],
        [0, 0, 100],
        [0, 0, 100],
        [360, 81, 100],
    ],
}
EFFECT_CHRISTMAS = {
    "custom": 0,
    "id": "bwTatyinOUajKrDwzMmqxxJdnInQUgvM",
    "brightness": 100,
    "name": "Christmas",
    "segments": [0],
    "expansion_strategy": 1,
    "enable": 1,
    "type": "random",
    "hue_range": [136, 146],
    "saturation_range": [90, 100],
    "brightness_range": [50, 100],
    "duration": 5000,
    "transition": 0,
    "init_states": [[136, 0, 100]],
    "fadeoff": 2000,
    "random_seed": 100,
    "backgrounds": [[136, 98, 75], [136, 0, 0], [350, 0, 100], [350, 97, 94]],
}
EFFECT_FLICKER = {
    "custom": 0,
    "id": "bCTItKETDFfrKANolgldxfgOakaarARs",
    "brightness": 100,
    "name": "Flicker",
    "segments": [1],
    "expansion_strategy": 1,
    "enable": 1,
    "type": "random",
    "hue_range": [30, 40],
    "saturation_range": [100, 100],
    "brightness_range": [50, 100],
    "duration": 0,
    "transition": 0,
    "transition_range": [375, 500],
    "init_states": [[30, 81, 80]],
}
EFFECT_HANUKKAH = {
    "custom": 0,
    "id": "CdLeIgiKcQrLKMINRPTMbylATulQewLD",
    "brightness": 100,
    "name": "Hanukkah",
    "segments": [1],
    "expansion_strategy": 1,
    "enable": 1,
    "type": "random",
    "hue_range": [200, 210],
    "saturation_range": [0, 100],
    "brightness_range": [50, 100],
    "duration": 1500,
    "transition": 0,
    "transition_range": [400, 500],
    "init_states": [[35, 81, 80]],
}
EFFECT_HAUNTED_MANSION = {
    "custom": 0,
    "id": "oJnFHsVQzFUTeIOBAhMRfVeujmSauhjJ",
    "brightness": 80,
    "name": "Haunted Mansion",
    "segments": [80],
    "expansion_strategy": 2,
    "enable": 1,
    "type": "random",
    "hue_range": [45, 45],
    "saturation_range": [10, 10],
    "brightness_range": [0, 80],
    "duration": 0,
    "transition": 0,
    "transition_range": [50, 1500],
    "init_states": [[45, 10, 100]],
    "fadeoff": 200,
    "random_seed": 1,
    "backgrounds": [[45, 10, 100]],
}
EFFECT_ICICLE = {
    "custom": 0,
    "id": "joqVjlaTsgzmuQQBAlHRkkPAqkBUiqeb",
    "brightness": 70,
    "name": "Icicle",
    "segments": [0],
    "expansion_strategy": 1,
    "enable": 1,
    "type": "sequence",
    "duration": 0,
    "transition": 400,
    "direction": 4,
    "spread": 3,
    "repeat_times": 0,
    "sequence": [
        [190, 100, 70],
        [190, 100, 70],
        [190, 30, 50],
        [190, 100, 70],
        [190, 100, 70],
    ],
}
EFFECT_LIGHTNING = {
    "custom": 0,
    "id": "ojqpUUxdGHoIugGPknrUcRoyJiItsjuE",
    "brightness": 100,
    "name": "Lightning",
    "segments": [7, 20, 23, 32, 34, 35, 49, 65, 66, 74, 80],
    "expansion_strategy": 1,
    "enable": 1,
    "type": "random",
    "hue_range": [240, 240],
    "saturation_range": [10, 11],
    "brightness_range": [90, 100],
    "duration": 0,
    "transition": 50,
    "init_states": [[240, 30, 100]],
    "fadeoff": 150,
    "random_seed": 600,
    "backgrounds": [[200, 100, 100], [200, 50, 10], [210, 10, 50], [240, 10, 0]],
}
EFFECT_OCEAN = {
    "custom": 0,
    "id": "oJjUMosgEMrdumfPANKbkFmBcAdEQsPy",
    "brightness": 30,
    "name": "Ocean",
    "segments": [0],
    "expansion_strategy": 1,
    "enable": 1,
    "type": "sequence",
    "duration": 0,
    "transition": 2000,
    "direction": 3,
    "spread": 16,
    "repeat_times": 0,
    "sequence": [[198, 84, 30], [198, 70, 30], [198, 10, 30]],
}
EFFECT_RAINBOW = {
    "custom": 0,
    "id": "izRhLCQNcDzIKdpMPqSTtBMuAIoreAuT",
    "brightness": 100,
    "name": "Rainbow",
    "segments": [0],
    "expansion_strategy": 1,
    "enable": 1,
    "type": "sequence",
    "duration": 0,
    "transition": 1500,
    "direction": 1,
    "spread": 12,
    "repeat_times": 0,
    "sequence": [[0, 100, 100], [100, 100, 100], [200, 100, 100], [300, 100, 100]],
}
EFFECT_RAINDROP = {
    "custom": 0,
    "id": "QbDFwiSFmLzQenUOPnJrsGqyIVrJrRsl",
    "brightness": 30,
    "name": "Raindrop",
    "segments": [0],
    "expansion_strategy": 1,
    "enable": 1,
    "type": "random",
    "hue_range": [200, 200],
    "saturation_range": [10, 20],
    "brightness_range": [10, 30],
    "duration": 0,
    "transition": 1000,
    "init_states": [[200, 40, 100]],
    "fadeoff": 1000,
    "random_seed": 24,
    "backgrounds": [[200, 40, 0]],
}
EFFECT_SPRING = {
    "custom": 0,
    "id": "URdUpEdQbnOOechDBPMkKrwhSupLyvAg",
    "brightness": 100,
    "name": "Spring",
    "segments": [0],
    "expansion_strategy": 1,
    "enable": 1,
    "type": "random",
    "hue_range": [0, 90],
    "saturation_range": [30, 100],
    "brightness_range": [90, 100],
    "duration": 600,
    "transition": 0,
    "transition_range": [2000, 6000],
    "init_states": [[80, 30, 100]],
    "fadeoff": 1000,
    "random_seed": 20,
    "backgrounds": [[130, 100, 40]],
}
EFFECT_VALENTINES = {
    "custom": 0,
    "id": "QglBhMShPHUAuxLqzNEefFrGiJwahOmz",
    "brightness": 100,
    "name": "Valentines",
    "segments": [0],
    "expansion_strategy": 1,
    "enable": 1,
    "type": "random",
    "hue_range": [340, 340],
    "saturation_range": [30, 40],
    "brightness_range": [90, 100],
    "duration": 600,
    "transition": 2000,
    "init_states": [[340, 30, 100]],
    "fadeoff": 3000,
    "random_seed": 100,
    "backgrounds": [[340, 20, 50], [20, 50, 50], [0, 100, 50]],
}

EFFECTS_LIST_V1 = [
    EFFECT_AURORA,
    EFFECT_BUBBLING_CAULDRON,
    EFFECT_CANDY_CANE,
    EFFECT_CHRISTMAS,
    EFFECT_FLICKER,
    EFFECT_HANUKKAH,
    EFFECT_HAUNTED_MANSION,
    EFFECT_ICICLE,
    EFFECT_LIGHTNING,
    EFFECT_OCEAN,
    EFFECT_RAINBOW,
    EFFECT_RAINDROP,
    EFFECT_SPRING,
    EFFECT_VALENTINES,
]

EFFECT_NAMES_V1: list[str] = [cast(str, effect["name"]) for effect in EFFECTS_LIST_V1]
EFFECT_MAPPING_V1 = {effect["name"]: effect for effect in EFFECTS_LIST_V1}



"""Implementation of temperature control module."""

from __future__ import annotations

import logging

from ...feature import Feature
from ...interfaces.thermostat import ThermostatState
from ..smartmodule import SmartModule

_LOGGER = logging.getLogger(__name__)


class TemperatureControl(SmartModule):
    """Implementation of temperature module."""

    REQUIRED_COMPONENT = "temp_control"

    def _initialize_features(self) -> None:
        """Initialize features after the initial update."""
        self._add_feature(
            Feature(
                self._device,
                id="target_temperature",
                name="Target temperature",
                container=self,
                attribute_getter="target_temperature",
                attribute_setter="set_target_temperature",
                range_getter="allowed_temperature_range",
                icon="mdi:thermometer",
                type=Feature.Type.Number,
                category=Feature.Category.Primary,
            )
        )
        # TODO: this might belong into its own module, temperature_correction?
        self._add_feature(
            Feature(
                self._device,
                id="temperature_offset",
                name="Temperature offset",
                container=self,
                attribute_getter="temperature_offset",
                attribute_setter="set_temperature_offset",
                range_getter=lambda: (-10, 10),
                type=Feature.Type.Number,
                category=Feature.Category.Config,
            )
        )
        self._add_feature(
            Feature(
                self._device,
                id="state",
                name="State",
                container=self,
                attribute_getter="state",
                attribute_setter="set_state",
                category=Feature.Category.Primary,
                type=Feature.Type.Switch,
            )
        )
        self._add_feature(
            Feature(
                self._device,
                id="thermostat_mode",
                name="Thermostat mode",
                container=self,
                attribute_getter="mode",
                category=Feature.Category.Primary,
                type=Feature.Type.Sensor,
            )
        )

    def query(self) -> dict:
        """Query to execute during the update cycle."""
        # Target temperature is contained in the main device info response.
        return {}

    @property
    def state(self) -> bool:
        """Return thermostat state."""
        return self._device.sys_info["frost_protection_on"] is False

    async def set_state(self, enabled: bool) -> dict:
        """Set thermostat state."""
        return await self.call("set_device_info", {"frost_protection_on": not enabled})

    @property
    def mode(self) -> ThermostatState:
        """Return thermostat state."""
        # If frost protection is enabled, the thermostat is off.
        if self._device.sys_info.get("frost_protection_on", False):
            return ThermostatState.Off

        states = self.states

        # If the states is empty, the device is idling
        if not states:
            return ThermostatState.Idle

        # Discard known extra states, and report on unknown extra states
        states.discard("low_battery")
        if len(states) > 1:
            _LOGGER.warning("Got multiple states: %s", states)

        # Return the first known state
        for state in ThermostatState:
            if state.value in states:
                return state

        _LOGGER.warning("Got unknown state: %s", states)
        return ThermostatState.Unknown

    @property
    def allowed_temperature_range(self) -> tuple[int, int]:
        """Return allowed temperature range."""
        return self.minimum_target_temperature, self.maximum_target_temperature

    @property
    def minimum_target_temperature(self) -> int:
        """Minimum available target temperature."""
        return self._device.sys_info["min_control_temp"]

    @property
    def maximum_target_temperature(self) -> int:
        """Minimum available target temperature."""
        return self._device.sys_info["max_control_temp"]

    @property
    def target_temperature(self) -> float:
        """Return target temperature."""
        return self._device.sys_info["target_temp"]

    @property
    def states(self) -> set:
        """Return thermostat states."""
        return set(self._device.sys_info["trv_states"])

    async def set_target_temperature(self, target: float) -> dict:
        """Set target temperature."""
        if (
            target < self.minimum_target_temperature
            or target > self.maximum_target_temperature
        ):
            raise ValueError(
                f"Invalid target temperature {target}, must be in range "
                f"[{self.minimum_target_temperature},{self.maximum_target_temperature}]"
            )

        payload = {"target_temp": target}
        # If the device has frost protection, we set it off to enable heating
        if "frost_protection_on" in self._device.sys_info:
            payload["frost_protection_on"] = False

        return await self.call("set_device_info", payload)

    @property
    def temperature_offset(self) -> int:
        """Return temperature offset."""
        return self._device.sys_info["temp_offset"]

    async def set_temperature_offset(self, offset: int) -> dict:
        """Set temperature offset."""
        if offset < -10 or offset > 10:
            raise ValueError("Temperature offset must be [-10, 10]")

        return await self.call("set_device_info", {"temp_offset": offset})



"""Module for cli light control commands."""

import asyncclick as click

from kasa import (
    Device,
    Module,
)
from kasa.iot import (
    IotBulb,
)

from .common import echo, error, pass_dev_or_child


@click.group()
@pass_dev_or_child
def light(dev) -> None:
    """Commands to control light settings."""


@light.command()
@click.argument("brightness", type=click.IntRange(0, 100), default=None, required=False)
@click.option("--transition", type=int, required=False)
@pass_dev_or_child
async def brightness(dev: Device, brightness: int, transition: int):
    """Get or set brightness."""
    if not (light := dev.modules.get(Module.Light)) or not light.has_feature(
        "brightness"
    ):
        error("This device does not support brightness.")
        return

    if brightness is None:
        echo(f"Brightness: {light.brightness}")
        return light.brightness
    else:
        echo(f"Setting brightness to {brightness}")
        return await light.set_brightness(brightness, transition=transition)


@light.command()
@click.argument(
    "temperature", type=click.IntRange(2500, 9000), default=None, required=False
)
@click.option("--transition", type=int, required=False)
@pass_dev_or_child
async def temperature(dev: Device, temperature: int, transition: int):
    """Get or set color temperature."""
    if not (light := dev.modules.get(Module.Light)) or not (
        color_temp_feat := light.get_feature("color_temp")
    ):
        error("Device does not support color temperature")
        return

    if temperature is None:
        echo(f"Color temperature: {light.color_temp}")
        valid_temperature_range = color_temp_feat.range
        if valid_temperature_range != (0, 0):
            echo("(min: {}, max: {})".format(*valid_temperature_range))
        else:
            echo(
                "Temperature range unknown, please open a github issue"
                f" or a pull request for model '{dev.model}'"
            )
        return color_temp_feat.range
    else:
        echo(f"Setting color temperature to {temperature}")
        return await light.set_color_temp(temperature, transition=transition)


@light.command()
@click.argument("effect", type=click.STRING, default=None, required=False)
@click.pass_context
@pass_dev_or_child
async def effect(dev: Device, ctx, effect):
    """Set an effect."""
    if not (light_effect := dev.modules.get(Module.LightEffect)):
        error("Device does not support effects")
        return
    if effect is None:
        echo(
            f"Light effect: {light_effect.effect}\n"
            + f"Available Effects: {light_effect.effect_list}"
        )
        return light_effect.effect

    if effect not in light_effect.effect_list:
        raise click.BadArgumentUsage(
            f"Effect must be one of: {light_effect.effect_list}", ctx
        )

    echo(f"Setting Effect: {effect}")
    return await light_effect.set_effect(effect)


@light.command()
@click.argument("h", type=click.IntRange(0, 360), default=None, required=False)
@click.argument("s", type=click.IntRange(0, 100), default=None, required=False)
@click.argument("v", type=click.IntRange(0, 100), default=None, required=False)
@click.option("--transition", type=int, required=False)
@click.pass_context
@pass_dev_or_child
async def hsv(dev: Device, ctx, h, s, v, transition):
    """Get or set color in HSV."""
    if not (light := dev.modules.get(Module.Light)) or not light.has_feature("hsv"):
        error("Device does not support colors")
        return

    if h is None and s is None and v is None:
        echo(f"Current HSV: {light.hsv}")
        return light.hsv
    elif s is None or v is None:
        raise click.BadArgumentUsage("Setting a color requires 3 values.", ctx)
    else:
        echo(f"Setting HSV: {h} {s} {v}")
        return await light.set_hsv(h, s, v, transition=transition)


@light.group(invoke_without_command=True)
@pass_dev_or_child
@click.pass_context
async def presets(ctx, dev):
    """List and modify bulb setting presets."""
    if ctx.invoked_subcommand is None:
        return await ctx.invoke(presets_list)


@presets.command(name="list")
@pass_dev_or_child
def presets_list(dev: Device):
    """List presets."""
    if not (light_preset := dev.modules.get(Module.LightPreset)):
        error("Device does not support light presets")
        return

    for idx, preset in enumerate(light_preset.preset_states_list):
        echo(
            f"[{idx}] Hue: {preset.hue or '':3}  "
            f"Saturation: {preset.saturation or '':3}  "
            f"Brightness/Value: {preset.brightness or '':3}  "
            f"Temp: {preset.color_temp or '':4}"
        )

    return light_preset.preset_states_list


@presets.command(name="modify")
@click.argument("index", type=int)
@click.option("--brightness", type=int, required=False, default=None)
@click.option("--hue", type=int, required=False, default=None)
@click.option("--saturation", type=int, required=False, default=None)
@click.option("--temperature", type=int, required=False, default=None)
@pass_dev_or_child
async def presets_modify(dev: Device, index, brightness, hue, saturation, temperature):
    """Modify a preset."""
    if not (light_preset := dev.modules.get(Module.LightPreset)):
        error("Device does not support light presets")
        return

    max_index = len(light_preset.preset_states_list) - 1
    if index > len(light_preset.preset_states_list) - 1:
        error(f"Invalid index, must be between 0 and {max_index}")
        return

    if all([val is None for val in {brightness, hue, saturation, temperature}]):
        error("Need to supply at least one option to modify.")
        return

    # Preset names have `Not set`` as the first value
    preset_name = light_preset.preset_list[index + 1]
    preset = light_preset.preset_states_list[index]

    echo(f"Preset {preset_name} currently: {preset}")

    if brightness is not None and preset.brightness is not None:
        preset.brightness = brightness
    if hue is not None and preset.hue is not None:
        preset.hue = hue
    if saturation is not None and preset.saturation is not None:
        preset.saturation = saturation
    if temperature is not None and preset.temperature is not None:
        preset.color_temp = temperature

    echo(f"Updating preset {preset_name} to: {preset}")

    return await light_preset.save_preset(preset_name, preset)


@light.command()
@pass_dev_or_child
@click.option("--type", type=click.Choice(["soft", "hard"], case_sensitive=False))
@click.option("--last", is_flag=True)
@click.option("--preset", type=int)
async def turn_on_behavior(dev: Device, type, last, preset):
    """Modify bulb turn-on behavior."""
    if dev.device_type is not Device.Type.Bulb or not isinstance(dev, IotBulb):
        error("Presets only supported on iot bulbs")
        return
    settings = await dev.get_turn_on_behavior()
    echo(f"Current turn on behavior: {settings}")

    # Return if we are not setting the value
    if not type and not last and not preset:
        return settings

    # If we are setting the value, the type has to be specified
    if (last or preset) and type is None:
        echo("To set the behavior, you need to define --type")
        return

    behavior = getattr(settings, type)

    if last:
        echo(f"Going to set {type} to last")
        behavior.preset = None
    elif preset is not None:
        echo(f"Going to set {type} to preset {preset}")
        behavior.preset = preset

    return await dev.set_turn_on_behavior(settings)



"""Module for time interface."""

from __future__ import annotations

from abc import ABC, abstractmethod
from datetime import datetime, tzinfo

from ..module import Module


class Time(Module, ABC):
    """Base class for tplink time module."""

    @property
    @abstractmethod
    def time(self) -> datetime:
        """Return timezone aware current device time."""

    @property
    @abstractmethod
    def timezone(self) -> tzinfo:
        """Return current timezone."""

    @abstractmethod
    async def set_time(self, dt: datetime) -> dict:
        """Set the device time."""



"""Implementation of vacuum speaker."""

from __future__ import annotations

import logging
from typing import Annotated

from ...feature import Feature
from ...module import FeatureAttribute
from ..smartmodule import SmartModule

_LOGGER = logging.getLogger(__name__)


class Speaker(SmartModule):
    """Implementation of vacuum speaker."""

    REQUIRED_COMPONENT = "speaker"

    def _initialize_features(self) -> None:
        """Initialize features."""
        self._add_feature(
            Feature(
                self._device,
                id="locate",
                name="Locate device",
                container=self,
                attribute_setter="locate",
                category=Feature.Category.Primary,
                type=Feature.Action,
            )
        )
        self._add_feature(
            Feature(
                self._device,
                id="volume",
                name="Volume",
                container=self,
                attribute_getter="volume",
                attribute_setter="set_volume",
                range_getter=lambda: (0, 100),
                category=Feature.Category.Config,
                type=Feature.Type.Number,
            )
        )

    def query(self) -> dict:
        """Query to execute during the update cycle."""
        return {
            "getVolume": None,
        }

    @property
    def volume(self) -> Annotated[str, FeatureAttribute()]:
        """Return volume."""
        return self.data["volume"]

    async def set_volume(self, volume: int) -> Annotated[dict, FeatureAttribute()]:
        """Set volume."""
        if volume < 0 or volume > 100:
            raise ValueError("Volume must be between 0 and 100")

        return await self.call("setVolume", {"volume": volume})

    async def locate(self) -> dict:
        """Play sound to locate the device."""
        return await self.call("playSelectAudio", {"audio_type": "seek_me"})



from __future__ import annotations

from datetime import timedelta

import pytest
from pytest_mock import MockerFixture

from kasa import Module
from kasa.smart import SmartDevice
from kasa.smart.modules.consumables import CONSUMABLE_METAS

from ...device_fixtures import get_parent_and_child_modules, parametrize

consumables = parametrize(
    "has consumables", component_filter="consumables", protocol_filter={"SMART"}
)


@consumables
@pytest.mark.parametrize(
    "consumable_name", [consumable.id for consumable in CONSUMABLE_METAS]
)
@pytest.mark.parametrize("postfix", ["used", "remaining"])
async def test_features(dev: SmartDevice, consumable_name: str, postfix: str):
    """Test that features are registered and work as expected."""
    consumables = next(get_parent_and_child_modules(dev, Module.Consumables))
    assert consumables is not None

    feature_name = f"{consumable_name}_{postfix}"

    feat = consumables._device.features[feature_name]
    assert isinstance(feat.value, timedelta)


@consumables
@pytest.mark.parametrize(
    ("consumable_name", "data_key"),
    [(consumable.id, consumable.data_key) for consumable in CONSUMABLE_METAS],
)
async def test_erase(
    dev: SmartDevice, mocker: MockerFixture, consumable_name: str, data_key: str
):
    """Test autocollection switch."""
    consumables = next(get_parent_and_child_modules(dev, Module.Consumables))
    call = mocker.spy(consumables, "call")

    feature_name = f"{consumable_name}_reset"
    feat = dev._features[feature_name]
    await feat.set_value(True)

    call.assert_called_with(
        "resetConsumablesTime", {"reset_list": [data_key.removesuffix("_time")]}
    )



"""Tests for smart camera devices."""

from __future__ import annotations

import pytest

from kasa import Device, Module
from kasa.smartcam.modules.alarm import (
    DURATION_MAX,
    DURATION_MIN,
    VOLUME_MAX,
    VOLUME_MIN,
)

from ...conftest import hub_smartcam


@hub_smartcam
async def test_alarm(dev: Device):
    """Test device alarm."""
    alarm = dev.modules.get(Module.Alarm)
    assert alarm

    original_duration = alarm.alarm_duration
    assert original_duration is not None
    original_volume = alarm.alarm_volume
    assert original_volume is not None
    original_sound = alarm.alarm_sound

    try:
        # test volume
        new_volume = original_volume - 1 if original_volume > 1 else original_volume + 1
        await alarm.set_alarm_volume(new_volume)  # type: ignore[arg-type]
        await dev.update()
        assert alarm.alarm_volume == new_volume

        # test duration
        new_duration = (
            original_duration - 1 if original_duration > 1 else original_duration + 1
        )
        await alarm.set_alarm_duration(new_duration)
        await dev.update()
        assert alarm.alarm_duration == new_duration

        # test start
        await alarm.play()
        await dev.update()
        assert alarm.active

        # test stop
        await alarm.stop()
        await dev.update()
        assert not alarm.active

        # test set sound
        new_sound = (
            alarm.alarm_sounds[0]
            if alarm.alarm_sound != alarm.alarm_sounds[0]
            else alarm.alarm_sounds[1]
        )
        await alarm.set_alarm_sound(new_sound)
        await dev.update()
        assert alarm.alarm_sound == new_sound

        # Test play parameters
        await alarm.play(
            duration=original_duration, volume=original_volume, sound=original_sound
        )
        await dev.update()
        assert alarm.active
        assert alarm.alarm_sound == original_sound
        assert alarm.alarm_duration == original_duration
        assert alarm.alarm_volume == original_volume
        await alarm.stop()
        await dev.update()
        assert not alarm.active

    finally:
        await alarm.set_alarm_volume(original_volume)
        await alarm.set_alarm_duration(original_duration)
        await alarm.set_alarm_sound(original_sound)
        await dev.update()


@hub_smartcam
async def test_alarm_invalid_setters(dev: Device):
    """Test device alarm invalid setter values."""
    alarm = dev.modules.get(Module.Alarm)
    assert alarm

    # test set sound invalid
    msg = f"sound must be one of {', '.join(alarm.alarm_sounds)}: foobar"
    with pytest.raises(ValueError, match=msg):
        await alarm.set_alarm_sound("foobar")

    # test volume invalid
    msg = f"volume must be between {VOLUME_MIN} and {VOLUME_MAX}"
    with pytest.raises(ValueError, match=msg):
        await alarm.set_alarm_volume(-3)

    # test duration invalid
    msg = f"duration must be between {DURATION_MIN} and {DURATION_MAX}"
    with pytest.raises(ValueError, match=msg):
        await alarm.set_alarm_duration(-3)


@hub_smartcam
async def test_alarm_features(dev: Device):
    """Test device alarm features."""
    alarm = dev.modules.get(Module.Alarm)
    assert alarm

    original_duration = alarm.alarm_duration
    assert original_duration is not None
    original_volume = alarm.alarm_volume
    assert original_volume is not None
    original_sound = alarm.alarm_sound

    try:
        # test volume
        new_volume = original_volume - 1 if original_volume > 1 else original_volume + 1
        feature = dev.features.get("alarm_volume")
        assert feature
        await feature.set_value(new_volume)  # type: ignore[arg-type]
        await dev.update()
        assert feature.value == new_volume

        # test duration
        feature = dev.features.get("alarm_duration")
        assert feature
        new_duration = (
            original_duration - 1 if original_duration > 1 else original_duration + 1
        )
        await feature.set_value(new_duration)
        await dev.update()
        assert feature.value == new_duration

        # test start
        feature = dev.features.get("test_alarm")
        assert feature
        await feature.set_value(None)
        await dev.update()
        feature = dev.features.get("alarm")
        assert feature
        assert feature.value is True

        # test stop
        feature = dev.features.get("stop_alarm")
        assert feature
        await feature.set_value(None)
        await dev.update()
        assert dev.features["alarm"].value is False

        # test set sound
        feature = dev.features.get("alarm_sound")
        assert feature
        new_sound = (
            alarm.alarm_sounds[0]
            if alarm.alarm_sound != alarm.alarm_sounds[0]
            else alarm.alarm_sounds[1]
        )
        await feature.set_value(new_sound)
        await alarm.set_alarm_sound(new_sound)
        await dev.update()
        assert feature.value == new_sound

    finally:
        await alarm.set_alarm_volume(original_volume)
        await alarm.set_alarm_duration(original_duration)
        await alarm.set_alarm_sound(original_sound)
        await dev.update()



"""Implementation of the antitheft module."""

from .rulemodule import RuleModule


class Antitheft(RuleModule):
    """Implementation of the antitheft module.

    This shares the functionality among other rule-based modules.
    """



"""Module for individual feature modules."""

from .ambientlight import AmbientLight
from .antitheft import Antitheft
from .cloud import Cloud
from .countdown import Countdown
from .dimmer import Dimmer
from .emeter import Emeter
from .led import Led
from .light import Light
from .lighteffect import LightEffect
from .lightpreset import IotLightPreset, LightPreset
from .motion import Motion
from .rulemodule import Rule, RuleModule
from .schedule import Schedule
from .time import Time
from .usage import Usage

__all__ = [
    "AmbientLight",
    "Antitheft",
    "Cloud",
    "Countdown",
    "Dimmer",
    "Emeter",
    "Led",
    "Light",
    "LightEffect",
    "LightPreset",
    "IotLightPreset",
    "Motion",
    "Rule",
    "RuleModule",
    "Schedule",
    "Time",
    "Usage",
]



"""Implementation of time module."""

from __future__ import annotations

from ...feature import Feature
from ..smartcammodule import SmartCamModule

DEFAULT_PAN_STEP = 30
DEFAULT_TILT_STEP = 10


class PanTilt(SmartCamModule):
    """Implementation of device_local_time."""

    REQUIRED_COMPONENT = "ptz"
    _pan_step = DEFAULT_PAN_STEP
    _tilt_step = DEFAULT_TILT_STEP

    def _initialize_features(self) -> None:
        """Initialize features after the initial update."""

        async def set_pan_step(value: int) -> None:
            self._pan_step = value

        async def set_tilt_step(value: int) -> None:
            self._tilt_step = value

        self._add_feature(
            Feature(
                self._device,
                "pan_right",
                "Pan right",
                container=self,
                attribute_setter=lambda: self.pan(self._pan_step * -1),
                type=Feature.Type.Action,
            )
        )
        self._add_feature(
            Feature(
                self._device,
                "pan_left",
                "Pan left",
                container=self,
                attribute_setter=lambda: self.pan(self._pan_step),
                type=Feature.Type.Action,
            )
        )
        self._add_feature(
            Feature(
                self._device,
                "pan_step",
                "Pan step",
                container=self,
                attribute_getter="_pan_step",
                attribute_setter=set_pan_step,
                type=Feature.Type.Number,
            )
        )
        self._add_feature(
            Feature(
                self._device,
                "tilt_up",
                "Tilt up",
                container=self,
                attribute_setter=lambda: self.tilt(self._tilt_step),
                type=Feature.Type.Action,
            )
        )
        self._add_feature(
            Feature(
                self._device,
                "tilt_down",
                "Tilt down",
                container=self,
                attribute_setter=lambda: self.tilt(self._tilt_step * -1),
                type=Feature.Type.Action,
            )
        )
        self._add_feature(
            Feature(
                self._device,
                "tilt_step",
                "Tilt step",
                container=self,
                attribute_getter="_tilt_step",
                attribute_setter=set_tilt_step,
                type=Feature.Type.Number,
            )
        )

    def query(self) -> dict:
        """Query to execute during the update cycle."""
        return {}

    async def pan(self, pan: int) -> dict:
        """Pan horizontally."""
        return await self.move(pan=pan, tilt=0)

    async def tilt(self, tilt: int) -> dict:
        """Tilt vertically."""
        return await self.move(pan=0, tilt=tilt)

    async def move(self, *, pan: int, tilt: int) -> dict:
        """Pan and tilt camera."""
        return await self._device._raw_query(
            {"do": {"motor": {"move": {"x_coord": str(pan), "y_coord": str(tilt)}}}}
        )



"""Implementation of trigger logs module."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Annotated

from mashumaro import DataClassDictMixin
from mashumaro.types import Alias

from ..smartmodule import SmartModule


@dataclass
class LogEntry(DataClassDictMixin):
    """Presentation of a single log entry."""

    id: int
    event_id: Annotated[str, Alias("eventId")]
    timestamp: int
    event: str


class TriggerLogs(SmartModule):
    """Implementation of trigger logs."""

    REQUIRED_COMPONENT = "trigger_log"
    MINIMUM_UPDATE_INTERVAL_SECS = 60 * 60

    def query(self) -> dict:
        """Query to execute during the update cycle."""
        return {"get_trigger_logs": {"start_id": 0}}

    @property
    def logs(self) -> list[LogEntry]:
        """Return logs."""
        return [LogEntry.from_dict(log) for log in self.data["logs"]]



"""Interact with TPLink Smart Home devices.

Once you have a device via :ref:`Discovery <discover_target>` or
:ref:`Connect <connect_target>` you can start interacting with a device.

>>> from kasa import Discover
>>>
>>> dev = await Discover.discover_single(
>>>     "127.0.0.2",
>>>     username="user@example.com",
>>>     password="great_password"
>>> )
>>>

Most devices can be turned on and off

>>> await dev.turn_on()
>>> await dev.update()
>>> print(dev.is_on)
True

>>> await dev.turn_off()
>>> await dev.update()
>>> print(dev.is_on)
False

All devices provide several informational properties:

>>> dev.alias
Bedroom Lamp Plug
>>> dev.model
HS110
>>> dev.rssi
-71
>>> dev.mac
50:C7:BF:00:00:00

Some information can also be changed programmatically:

>>> await dev.set_alias("new alias")
>>> await dev.update()
>>> dev.alias
new alias

Devices support different functionality that are exposed via
:ref:`modules <module_target>` that you can access via :attr:`~kasa.Device.modules`:

>>> for module_name in dev.modules:
>>>     print(module_name)
Energy
schedule
usage
anti_theft
Time
cloud
Led

>>> led_module = dev.modules["Led"]
>>> print(led_module.led)
False
>>> await led_module.set_led(True)
>>> await dev.update()
>>> print(led_module.led)
True

Individual pieces of functionality are also exposed via :ref:`features <feature_target>`
which you can access via :attr:`~kasa.Device.features` and will only be present if
they are supported.

Features are similar to modules in that they provide functionality that may or may
not be present.

Whereas modules group functionality into a common interface, features expose a single
function that may or may not be part of a module.

The advantage of features is that they have a simple common interface of `id`, `name`,
`value` and `set_value` so no need to learn the module API.

They are useful if you want write code that dynamically adapts as new features are
added to the API.

>>> for feature_name in dev.features:
>>>     print(feature_name)
state
rssi
on_since
reboot
current_consumption
consumption_today
consumption_this_month
consumption_total
voltage
current
cloud_connection
led

>>> led_feature = dev.features["led"]
>>> print(led_feature.value)
True
>>> await led_feature.set_value(False)
>>> await dev.update()
>>> print(led_feature.value)
False
"""

from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from collections.abc import Callable, Mapping, Sequence
from dataclasses import dataclass
from datetime import datetime, tzinfo
from typing import TYPE_CHECKING, Any, TypeAlias
from warnings import warn

from .credentials import Credentials as _Credentials
from .device_type import DeviceType
from .deviceconfig import (
    DeviceConfig,
    DeviceConnectionParameters,
    DeviceEncryptionType,
    DeviceFamily,
)
from .exceptions import KasaException
from .feature import Feature
from .module import Module
from .protocols import BaseProtocol, IotProtocol
from .transports import XorTransport

if TYPE_CHECKING:
    from .modulemapping import ModuleMapping, ModuleName


@dataclass
class WifiNetwork:
    """Wifi network container."""

    ssid: str
    key_type: int
    # These are available only on softaponboarding
    cipher_type: int | None = None
    bssid: str | None = None
    channel: int | None = None
    rssi: int | None = None

    # For SMART devices
    signal_level: int | None = None


_LOGGER = logging.getLogger(__name__)


@dataclass
class DeviceInfo:
    """Device Model Information."""

    short_name: str
    long_name: str
    brand: str
    device_family: str
    device_type: DeviceType
    hardware_version: str
    firmware_version: str
    firmware_build: str | None
    requires_auth: bool
    region: str | None


class Device(ABC):
    """Common device interface.

    Do not instantiate this class directly, instead get a device instance from
    :func:`Device.connect()`, :func:`Discover.discover()`
    or :func:`Discover.discover_single()`.
    """

    # All types required to create devices directly via connect are aliased here
    # to avoid consumers having to do multiple imports.

    #: The type of device
    Type: TypeAlias = DeviceType
    #: The credentials for authentication
    Credentials: TypeAlias = _Credentials
    #: Configuration for connecting to the device
    Config: TypeAlias = DeviceConfig
    #: The family of the device, e.g. SMART.KASASWITCH.
    Family: TypeAlias = DeviceFamily
    #: The encryption for the device, e.g. Klap or Aes
    EncryptionType: TypeAlias = DeviceEncryptionType
    #: The connection type for the device.
    ConnectionParameters: TypeAlias = DeviceConnectionParameters

    def __init__(
        self,
        host: str,
        *,
        config: DeviceConfig | None = None,
        protocol: BaseProtocol | None = None,
    ) -> None:
        """Create a new Device instance.

        :param str host: host name or IP address of the device
        :param DeviceConfig config: device configuration
        :param BaseProtocol protocol: protocol for communicating with the device
        """
        if config and protocol:
            protocol._transport._config = config
        self.protocol: BaseProtocol = protocol or IotProtocol(
            transport=XorTransport(config=config or DeviceConfig(host=host)),
        )
        self._last_update: dict[str, Any] = {}
        _LOGGER.debug("Initializing %s of type %s", host, type(self))
        self._device_type = DeviceType.Unknown
        # TODO: typing Any is just as using dict | None would require separate
        #       checks in accessors. the @updated_required decorator does not ensure
        #       mypy that these are not accessed incorrectly.
        self._discovery_info: dict[str, Any] | None = None

        self._features: dict[str, Feature] = {}
        self._parent: Device | None = None
        self._children: Mapping[str, Device] = {}

    @staticmethod
    async def connect(
        *,
        host: str | None = None,
        config: DeviceConfig | None = None,
    ) -> Device:
        """Connect to a single device by the given hostname or device configuration.

        This method avoids the UDP based discovery process and
        will connect directly to the device.

        It is generally preferred to avoid :func:`discover_single()` and
        use this function instead as it should perform better when
        the WiFi network is congested or the device is not responding
        to discovery requests.

        :param host: Hostname of device to query
        :param config: Connection parameters to ensure the correct protocol
            and connection options are used.
        :rtype: SmartDevice
        :return: Object for querying/controlling found device.
        """
        from .device_factory import connect  # pylint: disable=import-outside-toplevel

        return await connect(host=host, config=config)  # type: ignore[arg-type]

    @abstractmethod
    async def update(self, update_children: bool = True) -> None:
        """Update the device."""

    async def disconnect(self) -> None:
        """Disconnect and close any underlying connection resources."""
        await self.protocol.close()

    @property
    @abstractmethod
    def modules(self) -> ModuleMapping[Module]:
        """Return the device modules."""

    @property
    @abstractmethod
    def is_on(self) -> bool:
        """Return true if the device is on."""

    @property
    def is_off(self) -> bool:
        """Return True if device is off."""
        return not self.is_on

    @abstractmethod
    async def turn_on(self, **kwargs) -> dict:
        """Turn on the device."""

    @abstractmethod
    async def turn_off(self, **kwargs) -> dict:
        """Turn off the device."""

    @abstractmethod
    async def set_state(self, on: bool) -> dict:
        """Set the device state to *on*.

        This allows turning the device on and off.
        See also *turn_off* and *turn_on*.
        """

    @property
    def host(self) -> str:
        """The device host."""
        return self.protocol._transport._host

    @host.setter
    def host(self, value: str) -> None:
        """Set the device host.

        Generally used by discovery to set the hostname after ip discovery.
        """
        self.protocol._transport._host = value
        self.protocol._transport._config.host = value

    @property
    def port(self) -> int:
        """The device port."""
        return self.protocol._transport._port

    @property
    def credentials(self) -> _Credentials | None:
        """The device credentials."""
        return self.protocol._transport._credentials

    @property
    def credentials_hash(self) -> str | None:
        """The protocol specific hash of the credentials the device is using."""
        return self.protocol._transport.credentials_hash

    @property
    def device_type(self) -> DeviceType:
        """Return the device type."""
        return self._device_type

    @abstractmethod
    def update_from_discover_info(self, info: dict) -> None:
        """Update state from info from the discover call."""

    @property
    def config(self) -> DeviceConfig:
        """Return the device configuration."""
        return self.protocol.config

    @property
    @abstractmethod
    def model(self) -> str:
        """Returns the device model."""

    @property
    def region(self) -> str | None:
        """Returns the device region."""
        return self.device_info.region

    @property
    def device_info(self) -> DeviceInfo:
        """Return device info."""
        return self._get_device_info(self._last_update, self._discovery_info)

    @staticmethod
    @abstractmethod
    def _get_device_info(
        info: dict[str, Any], discovery_info: dict[str, Any] | None
    ) -> DeviceInfo:
        """Get device info."""

    @property
    @abstractmethod
    def alias(self) -> str | None:
        """Returns the device alias or nickname."""

    async def _raw_query(self, request: str | dict) -> dict:
        """Send a raw query to the device."""
        return await self.protocol.query(request=request)

    @property
    def parent(self) -> Device | None:
        """Return the parent on child devices."""
        return self._parent

    @property
    def children(self) -> Sequence[Device]:
        """Returns the child devices."""
        return list(self._children.values())

    def get_child_device(self, name_or_id: str) -> Device | None:
        """Return child device by its device_id or alias."""
        if name_or_id in self._children:
            return self._children[name_or_id]
        name_lower = name_or_id.lower()
        for child in self.children:
            if child.alias and child.alias.lower() == name_lower:
                return child
        return None

    @property
    @abstractmethod
    def sys_info(self) -> dict[str, Any]:
        """Returns the device info."""

    def get_plug_by_name(self, name: str) -> Device:
        """Return child device for the given name."""
        for p in self.children:
            if p.alias == name:
                return p

        raise KasaException(f"Device has no child with {name}")

    def get_plug_by_index(self, index: int) -> Device:
        """Return child device for the given index."""
        if index + 1 > len(self.children) or index < 0:
            raise KasaException(
                f"Invalid index {index}, device has {len(self.children)} plugs"
            )
        return self.children[index]

    @property
    @abstractmethod
    def time(self) -> datetime:
        """Return the time."""

    @property
    @abstractmethod
    def timezone(self) -> tzinfo:
        """Return the timezone and time_difference."""

    @property
    @abstractmethod
    def hw_info(self) -> dict:
        """Return hardware info for the device."""

    @property
    @abstractmethod
    def location(self) -> dict:
        """Return the device location."""

    @property
    @abstractmethod
    def rssi(self) -> int | None:
        """Return the rssi."""

    @property
    @abstractmethod
    def mac(self) -> str:
        """Return the mac formatted with colons."""

    @property
    @abstractmethod
    def device_id(self) -> str:
        """Return the device id."""

    @property
    @abstractmethod
    def internal_state(self) -> dict:
        """Return all the internal state data."""

    @property
    def state_information(self) -> dict[str, Any]:
        """Return available features and their values."""
        return {feat.name: feat.value for feat in self._features.values()}

    @property
    def features(self) -> dict[str, Feature]:
        """Return the list of supported features."""
        return self._features

    def _add_feature(self, feature: Feature) -> None:
        """Add a new feature to the device."""
        if feature.id in self._features:
            raise KasaException(f"Duplicate feature id {feature.id}")
        assert feature.id is not None  # TODO: hack for typing # noqa: S101
        self._features[feature.id] = feature

    @property
    @abstractmethod
    def has_emeter(self) -> bool:
        """Return if the device has emeter."""

    @property
    @abstractmethod
    def on_since(self) -> datetime | None:
        """Return the time that the device was turned on or None if turned off.

        This returns a cached value if the device reported value difference is under
        five seconds to avoid device-caused jitter.
        """

    @abstractmethod
    async def wifi_scan(self) -> list[WifiNetwork]:
        """Scan for available wifi networks."""

    @abstractmethod
    async def wifi_join(
        self, ssid: str, password: str, keytype: str = "wpa2_psk"
    ) -> dict:
        """Join the given wifi network."""

    @abstractmethod
    async def set_alias(self, alias: str) -> dict:
        """Set the device name (alias)."""

    @abstractmethod
    async def reboot(self, delay: int = 1) -> None:
        """Reboot the device.

        Note that giving a delay of zero causes this to block,
        as the device reboots immediately without responding to the call.
        """

    @abstractmethod
    async def factory_reset(self) -> None:
        """Reset device back to factory settings.

        Note, this does not downgrade the firmware.
        """

    def __repr__(self) -> str:
        update_needed = " - update() needed" if not self._last_update else ""
        if not self._last_update and not self._discovery_info:
            return f"<{self.device_type} at {self.host}{update_needed}>"
        return (
            f"<{self.device_type} at {self.host} -"
            f" {self.alias} ({self.model}){update_needed}>"
        )

    _deprecated_device_type_attributes = {
        # is_type
        "is_bulb": (None, DeviceType.Bulb),
        "is_dimmer": (None, DeviceType.Dimmer),
        "is_light_strip": (None, DeviceType.LightStrip),
        "is_plug": (None, DeviceType.Plug),
        "is_wallswitch": (None, DeviceType.WallSwitch),
        "is_strip": (None, DeviceType.Strip),
        "is_strip_socket": (None, DeviceType.StripSocket),
    }

    def _get_replacing_attr(
        self, module_name: ModuleName | None, *attrs: Any
    ) -> str | None:
        # If module name is None check self
        if not module_name:
            check = self
        elif (check := self.modules.get(module_name)) is None:
            return None

        for attr in attrs:
            # Use dir() as opposed to hasattr() to avoid raising exceptions
            # from properties
            if attr in dir(check):
                return attr

        return None

    def _get_deprecated_callable_attribute(self, name: str) -> Any | None:
        vals: dict[str, tuple[ModuleName, Callable[[Any], Any], str]] = {
            "is_dimmable": (
                Module.Light,
                lambda c: c.has_feature("brightness"),
                'light_module.has_feature("brightness")',
            ),
            "is_color": (
                Module.Light,
                lambda c: c.has_feature("hsv"),
                'light_module.has_feature("hsv")',
            ),
            "is_variable_color_temp": (
                Module.Light,
                lambda c: c.has_feature("color_temp"),
                'light_module.has_feature("color_temp")',
            ),
            "valid_temperature_range": (
                Module.Light,
                lambda c: c._deprecated_valid_temperature_range(),
                'minimum and maximum value of get_feature("color_temp")',
            ),
            "has_effects": (
                Module.Light,
                lambda c: Module.LightEffect in c._device.modules,
                "Module.LightEffect in device.modules",
            ),
        }
        if mod_call_msg := vals.get(name):
            mod, call, msg = mod_call_msg
            msg = f"{name} is deprecated, use: {msg} instead"
            warn(msg, DeprecationWarning, stacklevel=2)
            if (module := self.modules.get(mod)) is None:
                raise AttributeError(f"Device has no attribute {name!r}")
            return call(module)

        return None

    _deprecated_other_attributes = {
        # light attributes
        "brightness": (Module.Light, ["brightness"]),
        "set_brightness": (Module.Light, ["set_brightness"]),
        "hsv": (Module.Light, ["hsv"]),
        "set_hsv": (Module.Light, ["set_hsv"]),
        "color_temp": (Module.Light, ["color_temp"]),
        "set_color_temp": (Module.Light, ["set_color_temp"]),
        "_deprecated_set_light_state": (Module.Light, ["has_effects"]),
        # led attributes
        "led": (Module.Led, ["led"]),
        "set_led": (Module.Led, ["set_led"]),
        # light effect attributes
        # The return values for effect is a str instead of dict so the lightstrip
        # modules have a _deprecated method to return the value as before.
        "effect": (Module.LightEffect, ["_deprecated_effect", "effect"]),
        # The return values for effect_list includes the Off effect so the lightstrip
        # modules have a _deprecated method to return the values as before.
        "effect_list": (Module.LightEffect, ["_deprecated_effect_list", "effect_list"]),
        "set_effect": (Module.LightEffect, ["set_effect"]),
        "set_custom_effect": (Module.LightEffect, ["set_custom_effect"]),
        # light preset attributes
        "presets": (Module.LightPreset, ["_deprecated_presets", "preset_states_list"]),
        "save_preset": (Module.LightPreset, ["_deprecated_save_preset"]),
        # Emeter attribues
        "get_emeter_realtime": (Module.Energy, ["get_status"]),
        "emeter_realtime": (Module.Energy, ["status"]),
        "emeter_today": (Module.Energy, ["consumption_today"]),
        "emeter_this_month": (Module.Energy, ["consumption_this_month"]),
        "current_consumption": (Module.Energy, ["current_consumption"]),
        "get_emeter_daily": (Module.Energy, ["get_daily_stats"]),
        "get_emeter_monthly": (Module.Energy, ["get_monthly_stats"]),
        # Other attributes
        "supported_modules": (None, ["modules"]),
    }

    if not TYPE_CHECKING:

        def __getattr__(self, name: str) -> Any:
            # is_device_type
            if dep_device_type_attr := self._deprecated_device_type_attributes.get(
                name
            ):
                msg = f"{name} is deprecated, use device_type property instead"
                warn(msg, DeprecationWarning, stacklevel=2)
                return self.device_type == dep_device_type_attr[1]
            # callable
            if (result := self._get_deprecated_callable_attribute(name)) is not None:
                return result
            # Other deprecated attributes
            if (dep_attr := self._deprecated_other_attributes.get(name)) and (
                (replacing_attr := self._get_replacing_attr(dep_attr[0], *dep_attr[1]))
                is not None
            ):
                mod = dep_attr[0]
                dev_or_mod = self.modules[mod] if mod else self
                replacing = f"Module.{mod} in device.modules" if mod else replacing_attr
                msg = f"{name} is deprecated, use: {replacing} instead"
                warn(msg, DeprecationWarning, stacklevel=2)
                return getattr(dev_or_mod, replacing_attr)
            raise AttributeError(f"Device has no attribute {name!r}")



"""Implementation of the TP-Link SSL AES transport."""

from __future__ import annotations

import asyncio
import base64
import hashlib
import logging
import secrets
import ssl
from contextlib import suppress
from enum import Enum, auto
from typing import TYPE_CHECKING, Any, cast

from yarl import URL

from ..credentials import DEFAULT_CREDENTIALS, Credentials, get_default_credentials
from ..deviceconfig import DeviceConfig
from ..exceptions import (
    SMART_AUTHENTICATION_ERRORS,
    SMART_RETRYABLE_ERRORS,
    AuthenticationError,
    DeviceError,
    KasaException,
    SmartErrorCode,
    _RetryableError,
)
from ..httpclient import HttpClient
from ..json import dumps as json_dumps
from ..json import loads as json_loads
from . import AesEncyptionSession, BaseTransport

_LOGGER = logging.getLogger(__name__)


ONE_DAY_SECONDS = 86400
SESSION_EXPIRE_BUFFER_SECONDS = 60 * 20


def _sha256(payload: bytes) -> bytes:
    return hashlib.sha256(payload).digest()  # noqa: S324


def _md5_hash(payload: bytes) -> str:
    return hashlib.md5(payload).hexdigest().upper()  # noqa: S324


def _sha256_hash(payload: bytes) -> str:
    return hashlib.sha256(payload).hexdigest().upper()  # noqa: S324


class TransportState(Enum):
    """Enum for AES state."""

    HANDSHAKE_REQUIRED = auto()  # Handshake needed
    ESTABLISHED = auto()  # Ready to send requests


class SslAesTransport(BaseTransport):
    """Implementation of the AES encryption protocol.

    AES is the name used in device discovery for TP-Link's TAPO encryption
    protocol, sometimes used by newer firmware versions on kasa devices.
    """

    DEFAULT_PORT: int = 443
    COMMON_HEADERS = {
        "Content-Type": "application/json; charset=UTF-8",
        "requestByApp": "true",
        "Accept": "application/json",
        "Accept-Encoding": "gzip, deflate",
        "User-Agent": "Tapo CameraClient Android",
    }
    CIPHERS = ":".join(
        [
            "AES256-GCM-SHA384",
            "AES256-SHA256",
            "AES128-GCM-SHA256",
            "AES128-SHA256",
            "AES256-SHA",
        ]
    )
    DEFAULT_TIMEOUT = 10

    def __init__(
        self,
        *,
        config: DeviceConfig,
    ) -> None:
        super().__init__(config=config)

        self._login_version = config.connection_type.login_version
        if (
            not self._credentials or self._credentials.username is None
        ) and not self._credentials_hash:
            self._credentials = Credentials()
        self._default_credentials: Credentials = get_default_credentials(
            DEFAULT_CREDENTIALS["TAPOCAMERA"]
        )
        self._http_client: HttpClient = HttpClient(config)

        self._state = TransportState.HANDSHAKE_REQUIRED

        self._encryption_session: AesEncyptionSession | None = None
        self._session_expire_at: float | None = None

        self._host_port = f"{self._host}:{self._port}"
        self._app_url = URL(f"https://{self._host_port}")
        self._token_url: URL | None = None
        self._ssl_context: ssl.SSLContext | None = None
        ref = str(self._token_url) if self._token_url else str(self._app_url)
        self._headers = {
            **self.COMMON_HEADERS,
            "Host": self._host_port,
            "Referer": ref,
        }
        self._seq: int | None = None
        self._pwd_hash: str | None = None
        self._username: str | None = None
        self._password: str | None = None
        if self._credentials != Credentials() and self._credentials:
            self._username = self._credentials.username
            self._password = self._credentials.password
        elif self._credentials_hash:
            ch = json_loads(base64.b64decode(self._credentials_hash.encode()))
            self._password = ch["pwd"]
            self._username = ch["un"]
        self._local_nonce: str | None = None
        self._send_secure = True

        _LOGGER.debug("Created AES transport for %s", self._host)

    @property
    def default_port(self) -> int:
        """Default port for the transport."""
        if port := self._config.connection_type.http_port:
            return port
        return self.DEFAULT_PORT

    @staticmethod
    def _create_b64_credentials(credentials: Credentials) -> str:
        ch = {"un": credentials.username, "pwd": credentials.password}
        return base64.b64encode(json_dumps(ch).encode()).decode()

    @property
    def credentials_hash(self) -> str | None:
        """The hashed credentials used by the transport."""
        if self._credentials == Credentials():
            return None
        if not self._credentials and self._credentials_hash:
            return self._credentials_hash
        if (cred := self._credentials) and cred.password and cred.username:
            return self._create_b64_credentials(cred)
        return None

    def _get_response_error(self, resp_dict: Any) -> SmartErrorCode:
        error_code_raw = resp_dict.get("error_code")
        try:
            error_code = SmartErrorCode.from_int(error_code_raw)
        except ValueError:
            _LOGGER.warning(
                "Device %s received unknown error code: %s", self._host, error_code_raw
            )
            error_code = SmartErrorCode.INTERNAL_UNKNOWN_ERROR
        return error_code

    def _get_response_inner_error(self, resp_dict: Any) -> SmartErrorCode | None:
        # Device blocked errors have 'data' element at the root level, other inner
        # errors are inside 'result'
        error_code_raw = resp_dict.get("data", {}).get("code")

        if error_code_raw is None:
            error_code_raw = resp_dict.get("result", {}).get("data", {}).get("code")

        if error_code_raw is None:
            return None
        try:
            error_code = SmartErrorCode.from_int(error_code_raw)
        except ValueError:
            _LOGGER.warning(
                "Device %s received unknown error code: %s", self._host, error_code_raw
            )
            error_code = SmartErrorCode.INTERNAL_UNKNOWN_ERROR
        return error_code

    def _handle_response_error_code(self, resp_dict: Any, msg: str) -> None:
        error_code = self._get_response_error(resp_dict)
        if error_code is SmartErrorCode.SUCCESS:
            return
        msg = f"{msg}: {self._host}: {error_code.name}({error_code.value})"
        if error_code in SMART_RETRYABLE_ERRORS:
            raise _RetryableError(msg, error_code=error_code)
        if error_code in SMART_AUTHENTICATION_ERRORS:
            self._state = TransportState.HANDSHAKE_REQUIRED
            raise AuthenticationError(msg, error_code=error_code)
        raise DeviceError(msg, error_code=error_code)

    def _create_ssl_context(self) -> ssl.SSLContext:
        context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
        context.set_ciphers(self.CIPHERS)
        context.check_hostname = False
        context.verify_mode = ssl.CERT_NONE
        return context

    async def _get_ssl_context(self) -> ssl.SSLContext:
        if not self._ssl_context:
            loop = asyncio.get_running_loop()
            self._ssl_context = await loop.run_in_executor(
                None, self._create_ssl_context
            )
        return self._ssl_context

    async def send_secure_passthrough(self, request: str) -> dict[str, Any]:
        """Send encrypted message as passthrough."""
        if self._state is TransportState.ESTABLISHED and self._token_url:
            url = self._token_url
        else:
            url = self._app_url

        _LOGGER.debug(
            "Sending secure passthrough from %s",
            self._host,
        )
        encrypted_payload = self._encryption_session.encrypt(request.encode())  # type: ignore
        passthrough_request = {
            "method": "securePassthrough",
            "params": {"request": encrypted_payload.decode()},
        }
        passthrough_request_str = json_dumps(passthrough_request)
        if TYPE_CHECKING:
            assert self._pwd_hash
            assert self._local_nonce
            assert self._seq
        tag = self.generate_tag(
            passthrough_request_str, self._local_nonce, self._pwd_hash, self._seq
        )
        headers = {**self._headers, "Seq": str(self._seq), "Tapo_tag": tag}
        self._seq += 1
        status_code, resp_dict = await self._http_client.post(
            url,
            json=passthrough_request_str,
            headers=headers,
            ssl=await self._get_ssl_context(),
        )

        if TYPE_CHECKING:
            assert self._encryption_session is not None

        # Devices can respond with 500 if another session is created from
        # the same host. Decryption may not succeed after that
        if status_code == 500:
            msg = (
                f"Device {self._host} replied with status 500 after handshake, "
                f"response: "
            )
            decrypted = None
            if isinstance(resp_dict, dict) and (
                response := resp_dict.get("result", {}).get("response")
            ):
                with suppress(Exception):
                    decrypted = self._encryption_session.decrypt(response.encode())

            if decrypted:
                msg += decrypted
            else:
                msg += str(resp_dict)

            _LOGGER.debug(msg)
            raise _RetryableError(msg)

        if status_code != 200:
            raise KasaException(
                f"{self._host} responded with an unexpected "
                + f"status code {status_code} to passthrough"
            )

        self._handle_response_error_code(
            resp_dict, "Error sending secure_passthrough message"
        )

        if TYPE_CHECKING:
            resp_dict = cast(dict[str, Any], resp_dict)

        if "result" in resp_dict and "response" in resp_dict["result"]:
            raw_response: str = resp_dict["result"]["response"]
        else:
            # Tapo Cameras respond unencrypted to single requests.
            return resp_dict

        try:
            response = self._encryption_session.decrypt(raw_response.encode())
            ret_val = json_loads(response)
        except Exception as ex:
            try:
                ret_val = json_loads(raw_response)
                _LOGGER.debug(
                    "Received unencrypted response over secure passthrough from %s",
                    self._host,
                )
            except Exception:
                raise KasaException(
                    f"Unable to decrypt response from {self._host}, "
                    + f"error: {ex}, response: {raw_response}",
                    ex,
                ) from ex
        return ret_val  # type: ignore[return-value]

    async def send_unencrypted(self, request: str) -> dict[str, Any]:
        """Send encrypted message as passthrough."""
        url = cast(URL, self._token_url)

        _LOGGER.debug(
            "Sending unencrypted to %s",
            self._host,
        )

        status_code, resp_dict = await self._http_client.post(
            url,
            json=request,
            headers=self._headers,
            ssl=await self._get_ssl_context(),
        )

        if status_code != 200:
            raise KasaException(
                f"{self._host} responded with an unexpected "
                + f"status code {status_code} to unencrypted send"
            )

        self._handle_response_error_code(resp_dict, "Error sending message")

        if TYPE_CHECKING:
            resp_dict = cast(dict[str, Any], resp_dict)
        return resp_dict

    @staticmethod
    def generate_confirm_hash(
        local_nonce: str, server_nonce: str, pwd_hash: str
    ) -> str:
        """Generate an auth hash for the protocol on the supplied credentials."""
        expected_confirm_bytes = _sha256_hash(
            local_nonce.encode() + pwd_hash.encode() + server_nonce.encode()
        )
        return expected_confirm_bytes + server_nonce + local_nonce

    @staticmethod
    def generate_digest_password(
        local_nonce: str, server_nonce: str, pwd_hash: str
    ) -> str:
        """Generate an auth hash for the protocol on the supplied credentials."""
        digest_password_hash = _sha256_hash(
            pwd_hash.encode() + local_nonce.encode() + server_nonce.encode()
        )
        return (
            digest_password_hash.encode() + local_nonce.encode() + server_nonce.encode()
        ).decode()

    @staticmethod
    def generate_encryption_token(
        token_type: str, local_nonce: str, server_nonce: str, pwd_hash: str
    ) -> bytes:
        """Generate encryption token."""
        hashedKey = _sha256_hash(
            local_nonce.encode() + pwd_hash.encode() + server_nonce.encode()
        )
        return _sha256(
            token_type.encode()
            + local_nonce.encode()
            + server_nonce.encode()
            + hashedKey.encode()
        )[:16]

    @staticmethod
    def generate_tag(request: str, local_nonce: str, pwd_hash: str, seq: int) -> str:
        """Generate the tag header from the request for the header."""
        pwd_nonce_hash = _sha256_hash(pwd_hash.encode() + local_nonce.encode())
        tag = _sha256_hash(
            pwd_nonce_hash.encode() + request.encode() + str(seq).encode()
        )
        return tag

    async def perform_handshake(self) -> None:
        """Perform the handshake."""
        result = await self.perform_handshake1()
        if result:
            local_nonce, server_nonce, pwd_hash = result
            await self.perform_handshake2(local_nonce, server_nonce, pwd_hash)

    async def try_perform_less_secure_login(self, username: str, password: str) -> bool:
        """Perform the md5 login."""
        _LOGGER.debug("Performing less secure login...")

        pwd_hash = _md5_hash(password.encode())
        body = {
            "method": "login",
            "params": {
                "hashed": True,
                "password": pwd_hash,
                "username": username,
            },
        }

        status_code, resp_dict = await self._http_client.post(
            self._app_url,
            json=body,
            headers=self._headers,
            ssl=await self._get_ssl_context(),
        )
        if status_code != 200:
            raise KasaException(
                f"{self._host} responded with an unexpected "
                + f"status code {status_code} to login"
            )
        resp_dict = cast(dict, resp_dict)
        if resp_dict.get("error_code") == 0 and (
            stok := resp_dict.get("result", {}).get("stok")
        ):
            _LOGGER.debug(
                "Succesfully logged in to %s with less secure passthrough", self._host
            )
            self._send_secure = False
            self._token_url = URL(f"{str(self._app_url)}/stok={stok}/ds")
            self._pwd_hash = pwd_hash
            return True

        _LOGGER.debug("Unable to log in to %s with less secure login", self._host)
        return False

    async def perform_handshake2(
        self, local_nonce: str, server_nonce: str, pwd_hash: str
    ) -> None:
        """Perform the handshake."""
        _LOGGER.debug("Performing handshake2 ...")
        digest_password = self.generate_digest_password(
            local_nonce, server_nonce, pwd_hash
        )
        body = {
            "method": "login",
            "params": {
                "cnonce": local_nonce,
                "encrypt_type": "3",
                "digest_passwd": digest_password,
                "username": self._username,
            },
        }
        http_client = self._http_client
        status_code, resp_dict = await http_client.post(
            self._app_url,
            json=body,
            headers=self._headers,
            ssl=await self._get_ssl_context(),
        )
        if status_code != 200:
            raise KasaException(
                f"{self._host} responded with an unexpected "
                + f"status code {status_code} to handshake2"
            )
        resp_dict = cast(dict, resp_dict)
        if (
            error_code := self._get_response_error(resp_dict)
        ) and error_code is SmartErrorCode.INVALID_NONCE:
            raise AuthenticationError(
                f"Invalid password hash in handshake2 for {self._host}"
            )

        self._handle_response_error_code(resp_dict, "Error in handshake2")

        self._seq = resp_dict["result"]["start_seq"]
        stok = resp_dict["result"]["stok"]
        self._token_url = URL(f"{str(self._app_url)}/stok={stok}/ds")
        self._pwd_hash = pwd_hash
        self._local_nonce = local_nonce
        lsk = self.generate_encryption_token("lsk", local_nonce, server_nonce, pwd_hash)
        ivb = self.generate_encryption_token("ivb", local_nonce, server_nonce, pwd_hash)
        self._encryption_session = AesEncyptionSession(lsk, ivb)
        self._state = TransportState.ESTABLISHED
        _LOGGER.debug("Handshake2 complete ...")

    def _pwd_to_hash(self) -> str:
        """Return the password to hash."""
        if self._credentials and self._credentials != Credentials():
            return self._credentials.password

        if self._username and self._password:
            return self._password

        return self._default_credentials.password

    def _is_less_secure_login(self, resp_dict: dict[str, Any]) -> bool:
        result = (
            self._get_response_error(resp_dict) is SmartErrorCode.SESSION_EXPIRED
            and (data := resp_dict.get("result", {}).get("data", {}))
            and (encrypt_type := data.get("encrypt_type"))
            and (encrypt_type != ["3"])
        )
        if result:
            _LOGGER.debug(
                "Received encrypt_type %s for %s, trying less secure login",
                encrypt_type,
                self._host,
            )
        return result

    async def perform_handshake1(self) -> tuple[str, str, str] | None:
        """Perform the handshake1."""
        resp_dict = None
        if self._username:
            local_nonce = secrets.token_bytes(8).hex().upper()
            resp_dict = await self.try_send_handshake1(self._username, local_nonce)

        if (
            resp_dict
            and self._is_less_secure_login(resp_dict)
            and self._get_response_inner_error(resp_dict)
            is not SmartErrorCode.BAD_USERNAME
            and await self.try_perform_less_secure_login(
                cast(str, self._username), self._pwd_to_hash()
            )
        ):
            self._state = TransportState.ESTABLISHED
            return None

        # Try the default username. If it fails raise the original error_code
        if (
            not resp_dict
            or (error_code := self._get_response_error(resp_dict))
            is not SmartErrorCode.INVALID_NONCE
            or "nonce" not in resp_dict["result"].get("data", {})
        ):
            _LOGGER.debug("Trying default credentials to %s", self._host)
            local_nonce = secrets.token_bytes(8).hex().upper()
            default_resp_dict = await self.try_send_handshake1(
                self._default_credentials.username, local_nonce
            )
            # INVALID_NONCE means device should perform secure login
            if (
                default_error_code := self._get_response_error(default_resp_dict)
            ) is SmartErrorCode.INVALID_NONCE and "nonce" in default_resp_dict[
                "result"
            ].get("data", {}):
                _LOGGER.debug("Connected to %s with default username", self._host)
                self._username = self._default_credentials.username
                error_code = default_error_code
                resp_dict = default_resp_dict
            # Otherwise could be less secure login
            elif self._is_less_secure_login(
                default_resp_dict
            ) and await self.try_perform_less_secure_login(
                self._default_credentials.username, self._pwd_to_hash()
            ):
                self._username = self._default_credentials.username
                self._state = TransportState.ESTABLISHED
                return None

        # If the default login worked it's ok not to provide credentials but if
        # it didn't raise auth error here.
        if not self._username:
            raise AuthenticationError(
                f"Credentials must be supplied to connect to {self._host}"
            )

        # Device responds with INVALID_NONCE and a "nonce" to indicate ready
        # for secure login. Otherwise error.
        if error_code is not SmartErrorCode.INVALID_NONCE or (
            resp_dict and "nonce" not in resp_dict.get("result", {}).get("data", {})
        ):
            if (
                resp_dict
                and self._get_response_inner_error(resp_dict)
                is SmartErrorCode.DEVICE_BLOCKED
            ):
                sec_left = resp_dict.get("data", {}).get("sec_left")
                msg = "Device blocked" + (
                    f" for {sec_left} seconds" if sec_left else ""
                )
                raise DeviceError(msg, error_code=SmartErrorCode.DEVICE_BLOCKED)

            raise AuthenticationError(f"Error trying handshake1: {resp_dict}")

        if TYPE_CHECKING:
            resp_dict = cast(dict[str, Any], resp_dict)

        server_nonce = resp_dict["result"]["data"]["nonce"]
        device_confirm = resp_dict["result"]["data"]["device_confirm"]

        pwd_hash = _sha256_hash(self._pwd_to_hash().encode())

        expected_confirm_sha256 = self.generate_confirm_hash(
            local_nonce, server_nonce, pwd_hash
        )
        if device_confirm == expected_confirm_sha256:
            _LOGGER.debug("Credentials match")
            return local_nonce, server_nonce, pwd_hash

        if TYPE_CHECKING:
            assert self._credentials
            assert self._credentials.password

        pwd_hash = _md5_hash(self._pwd_to_hash().encode())

        expected_confirm_md5 = self.generate_confirm_hash(
            local_nonce, server_nonce, pwd_hash
        )
        if device_confirm == expected_confirm_md5:
            _LOGGER.debug("Credentials match")
            return local_nonce, server_nonce, pwd_hash

        msg = (
            f"Device response did not match our challenge on ip {self._host}, "
            f"check that your e-mail and password (both case-sensitive) are correct. "
        )
        _LOGGER.debug(msg)

        raise AuthenticationError(msg)

    async def try_send_handshake1(self, username: str, local_nonce: str) -> dict:
        """Perform the handshake."""
        _LOGGER.debug("Sending handshake1...")

        body = {
            "method": "login",
            "params": {
                "cnonce": local_nonce,
                "encrypt_type": "3",
                "username": username,
            },
        }
        http_client = self._http_client

        status_code, resp_dict = await http_client.post(
            self._app_url,
            json=body,
            headers=self._headers,
            ssl=await self._get_ssl_context(),
        )

        _LOGGER.debug("Device responded with status %s: %s", status_code, resp_dict)

        if status_code != 200:
            raise KasaException(
                f"{self._host} responded with an unexpected "
                + f"status code {status_code} to handshake1"
            )

        return cast(dict, resp_dict)

    async def send(self, request: str) -> dict[str, Any]:
        """Send the request."""
        if self._state is TransportState.HANDSHAKE_REQUIRED:
            await self.perform_handshake()

        if self._send_secure:
            return await self.send_secure_passthrough(request)

        return await self.send_unencrypted(request)

    async def close(self) -> None:
        """Close the http client and reset internal state."""
        await self.reset()
        await self._http_client.close()

    async def reset(self) -> None:
        """Reset internal handshake state."""
        self._state = TransportState.HANDSHAKE_REQUIRED
        self._encryption_session = None
        self._seq = 0
        self._pwd_hash = None
        self._local_nonce = None



from __future__ import annotations

import logging
import secrets
from contextlib import nullcontext as does_not_raise
from json import dumps as json_dumps
from json import loads as json_loads
from typing import Any

import aiohttp
import pytest
from yarl import URL

from kasa.credentials import DEFAULT_CREDENTIALS, Credentials, get_default_credentials
from kasa.deviceconfig import DeviceConfig
from kasa.exceptions import (
    AuthenticationError,
    DeviceError,
    KasaException,
    SmartErrorCode,
    _RetryableError,
)
from kasa.httpclient import HttpClient
from kasa.transports.aestransport import AesEncyptionSession
from kasa.transports.sslaestransport import (
    SslAesTransport,
    TransportState,
    _md5_hash,
    _sha256_hash,
)

# Transport tests are not designed for real devices
# SslAesTransport use a socket to get it's own ip address
pytestmark = [pytest.mark.requires_dummy, pytest.mark.enable_socket]

MOCK_ADMIN_USER = get_default_credentials(DEFAULT_CREDENTIALS["TAPOCAMERA"]).username
MOCK_PWD = "correct_pwd"  # noqa: S105
MOCK_USER = "mock@example.com"
MOCK_STOCK = "abcdefghijklmnopqrstuvwxyz1234)("
MOCK_UNENCRYPTED_PASSTHROUGH_STOK = "32charLowerCaseHexStok"


@pytest.mark.parametrize(
    (
        "status_code",
        "username",
        "password",
        "wants_default_user",
        "digest_password_fail",
        "expectation",
    ),
    [
        pytest.param(
            200, MOCK_USER, MOCK_PWD, False, False, does_not_raise(), id="success"
        ),
        pytest.param(
            200,
            MOCK_USER,
            MOCK_PWD,
            True,
            False,
            does_not_raise(),
            id="success-default",
        ),
        pytest.param(
            400,
            MOCK_USER,
            MOCK_PWD,
            False,
            False,
            pytest.raises(KasaException),
            id="400 error",
        ),
        pytest.param(
            200,
            "foobar",
            MOCK_PWD,
            False,
            False,
            pytest.raises(AuthenticationError),
            id="bad-username",
        ),
        pytest.param(
            200,
            MOCK_USER,
            "barfoo",
            False,
            False,
            pytest.raises(AuthenticationError),
            id="bad-password",
        ),
        pytest.param(
            200,
            MOCK_USER,
            MOCK_PWD,
            False,
            True,
            pytest.raises(AuthenticationError),
            id="bad-password-digest",
        ),
    ],
)
async def test_handshake(
    mocker,
    status_code,
    username,
    password,
    wants_default_user,
    digest_password_fail,
    expectation,
):
    host = "127.0.0.1"
    mock_ssl_aes_device = MockSslAesDevice(
        host,
        status_code=status_code,
        want_default_username=wants_default_user,
        digest_password_fail=digest_password_fail,
    )
    mocker.patch.object(
        aiohttp.ClientSession, "post", side_effect=mock_ssl_aes_device.post
    )

    transport = SslAesTransport(
        config=DeviceConfig(host, credentials=Credentials(username, password))
    )

    assert transport._encryption_session is None
    assert transport._state is TransportState.HANDSHAKE_REQUIRED
    with expectation:
        await transport.perform_handshake()
        assert transport._encryption_session is not None
        assert transport._state is TransportState.ESTABLISHED


@pytest.mark.parametrize(
    ("wants_default_user"),
    [pytest.param(False, id="username"), pytest.param(True, id="default")],
)
async def test_credentials_hash(mocker, wants_default_user):
    host = "127.0.0.1"
    mock_ssl_aes_device = MockSslAesDevice(
        host, want_default_username=wants_default_user
    )
    mocker.patch.object(
        aiohttp.ClientSession, "post", side_effect=mock_ssl_aes_device.post
    )
    creds = Credentials(MOCK_USER, MOCK_PWD)
    creds_hash = SslAesTransport._create_b64_credentials(creds)

    # Test with credentials input
    transport = SslAesTransport(config=DeviceConfig(host, credentials=creds))
    assert transport.credentials_hash == creds_hash
    await transport.perform_handshake()
    assert transport.credentials_hash == creds_hash

    # Test with credentials_hash input
    transport = SslAesTransport(config=DeviceConfig(host, credentials_hash=creds_hash))
    mock_ssl_aes_device.handshake1_complete = False
    assert transport.credentials_hash == creds_hash
    await transport.perform_handshake()
    assert transport.credentials_hash == creds_hash


async def test_send(mocker):
    host = "127.0.0.1"
    mock_ssl_aes_device = MockSslAesDevice(host, want_default_username=False)
    mocker.patch.object(
        aiohttp.ClientSession, "post", side_effect=mock_ssl_aes_device.post
    )

    transport = SslAesTransport(
        config=DeviceConfig(host, credentials=Credentials(MOCK_USER, MOCK_PWD))
    )
    request = {
        "method": "getDeviceInfo",
        "params": None,
    }

    res = await transport.send(json_dumps(request))
    assert "result" in res


@pytest.mark.xdist_group(name="caplog")
async def test_unencrypted_response(mocker, caplog):
    host = "127.0.0.1"
    mock_ssl_aes_device = MockSslAesDevice(host, do_not_encrypt_response=True)
    mocker.patch.object(
        aiohttp.ClientSession, "post", side_effect=mock_ssl_aes_device.post
    )

    transport = SslAesTransport(
        config=DeviceConfig(host, credentials=Credentials(MOCK_USER, MOCK_PWD))
    )

    request = {
        "method": "getDeviceInfo",
        "params": None,
    }
    caplog.set_level(logging.DEBUG)
    res = await transport.send(json_dumps(request))
    assert "result" in res
    assert (
        "Received unencrypted response over secure passthrough from 127.0.0.1"
        in caplog.text
    )


@pytest.mark.parametrize(("want_default"), [True, False])
@pytest.mark.xdist_group(name="caplog")
async def test_unencrypted_passthrough(mocker, caplog, want_default):
    host = "127.0.0.1"
    mock_ssl_aes_device = MockSslAesDevice(
        host, unencrypted_passthrough=True, want_default_username=want_default
    )
    mocker.patch.object(
        aiohttp.ClientSession, "post", side_effect=mock_ssl_aes_device.post
    )

    transport = SslAesTransport(
        config=DeviceConfig(host, credentials=Credentials(MOCK_USER, MOCK_PWD))
    )

    request = {
        "method": "getDeviceInfo",
        "params": None,
    }
    caplog.set_level(logging.DEBUG)
    res = await transport.send(json_dumps(request))
    assert "result" in res
    assert (
        f"Succesfully logged in to {host} with less secure passthrough" in caplog.text
    )


@pytest.mark.parametrize(("want_default"), [True, False])
@pytest.mark.xdist_group(name="caplog")
async def test_unencrypted_passthrough_errors(mocker, caplog, want_default):
    host = "127.0.0.1"
    request = {
        "method": "getDeviceInfo",
        "params": None,
    }
    transport = SslAesTransport(
        config=DeviceConfig(host, credentials=Credentials(MOCK_USER, MOCK_PWD))
    )
    caplog.set_level(logging.DEBUG)

    # Test bad password
    mock_ssl_aes_device = MockSslAesDevice(
        host,
        unencrypted_passthrough=True,
        want_default_username=want_default,
        digest_password_fail=True,
    )
    mocker.patch.object(
        aiohttp.ClientSession, "post", side_effect=mock_ssl_aes_device.post
    )

    msg = f"Unable to log in to {host} with less secure login"
    with pytest.raises(AuthenticationError):
        await transport.send(json_dumps(request))

    assert msg in caplog.text

    # Test bad status code in handshake
    mock_ssl_aes_device = MockSslAesDevice(
        host,
        unencrypted_passthrough=True,
        want_default_username=want_default,
        status_code=401,
    )
    mocker.patch.object(
        aiohttp.ClientSession, "post", side_effect=mock_ssl_aes_device.post
    )

    msg = f"{host} responded with an unexpected status code 401 to handshake1"
    with pytest.raises(KasaException, match=msg):
        await transport.send(json_dumps(request))

    # Test bad status code in login
    mock_ssl_aes_device = MockSslAesDevice(
        host,
        unencrypted_passthrough=True,
        want_default_username=want_default,
        status_code_list=[200, 401],
    )
    mocker.patch.object(
        aiohttp.ClientSession, "post", side_effect=mock_ssl_aes_device.post
    )

    msg = f"{host} responded with an unexpected status code 401 to login"
    with pytest.raises(KasaException, match=msg):
        await transport.send(json_dumps(request))

    # Test bad status code in send
    mock_ssl_aes_device = MockSslAesDevice(
        host,
        unencrypted_passthrough=True,
        want_default_username=want_default,
        status_code_list=[200, 200, 401],
    )
    mocker.patch.object(
        aiohttp.ClientSession, "post", side_effect=mock_ssl_aes_device.post
    )

    msg = f"{host} responded with an unexpected status code 401 to unencrypted send"
    with pytest.raises(KasaException, match=msg):
        await transport.send(json_dumps(request))

    # Test error code in send response
    mock_ssl_aes_device = MockSslAesDevice(
        host,
        unencrypted_passthrough=True,
        want_default_username=want_default,
        send_error_code=SmartErrorCode.BAD_USERNAME.value,
    )
    mocker.patch.object(
        aiohttp.ClientSession, "post", side_effect=mock_ssl_aes_device.post
    )

    msg = f"Error sending message: {host}:"
    with pytest.raises(KasaException, match=msg):
        await transport.send(json_dumps(request))


async def test_device_blocked_response(mocker):
    host = "127.0.0.1"
    mock_ssl_aes_device = MockSslAesDevice(host, device_blocked=True)
    mocker.patch.object(
        aiohttp.ClientSession, "post", side_effect=mock_ssl_aes_device.post
    )

    transport = SslAesTransport(
        config=DeviceConfig(host, credentials=Credentials(MOCK_USER, MOCK_PWD))
    )
    msg = "Device blocked for 1685 seconds"

    with pytest.raises(DeviceError, match=msg):
        await transport.perform_handshake()


@pytest.mark.parametrize(
    ("response", "expected_msg"),
    [
        pytest.param(
            {"error_code": -1, "msg": "Check tapo tag failed"},
            '{"error_code": -1, "msg": "Check tapo tag failed"}',
            id="can-decrypt",
        ),
        pytest.param(
            b"12345678",
            str({"result": {"response": "12345678"}, "error_code": 0}),
            id="cannot-decrypt",
        ),
    ],
)
async def test_device_500_error(mocker, response, expected_msg):
    """Test 500 error raises retryable exception."""
    host = "127.0.0.1"
    mock_ssl_aes_device = MockSslAesDevice(host)
    mocker.patch.object(
        aiohttp.ClientSession, "post", side_effect=mock_ssl_aes_device.post
    )

    transport = SslAesTransport(
        config=DeviceConfig(host, credentials=Credentials(MOCK_USER, MOCK_PWD))
    )

    request = {
        "method": "getDeviceInfo",
        "params": None,
    }

    await transport.perform_handshake()

    mock_ssl_aes_device.put_next_response(response)
    mock_ssl_aes_device.status_code = 500

    msg = f"Device 127.0.0.1 replied with status 500 after handshake, response: {expected_msg}"
    with pytest.raises(_RetryableError, match=msg):
        await transport.send(json_dumps(request))


async def test_port_override():
    """Test that port override sets the app_url."""
    host = "127.0.0.1"
    port_override = 12345
    config = DeviceConfig(
        host, credentials=Credentials("foo", "bar"), port_override=port_override
    )
    transport = SslAesTransport(config=config)

    assert str(transport._app_url) == f"https://127.0.0.1:{port_override}"


class MockSslAesDevice:
    BAD_USER_RESP = {
        "error_code": SmartErrorCode.SESSION_EXPIRED.value,
        "result": {
            "data": {
                "code": -60502,
            }
        },
    }

    BAD_PWD_RESP = {
        "error_code": SmartErrorCode.INVALID_NONCE.value,
        "result": {
            "data": {
                "code": SmartErrorCode.SESSION_EXPIRED.value,
                "encrypt_type": ["3"],
                "key": "Someb64keyWithUnknownPurpose",
                "nonce": "1234567890ABCDEF",  # Whatever the original nonce was
                "device_confirm": "",
            }
        },
    }

    DEVICE_BLOCKED_RESP = {
        "data": {"code": SmartErrorCode.DEVICE_BLOCKED.value, "sec_left": 1685},
        "error_code": SmartErrorCode.SESSION_EXPIRED.value,
    }

    UNENCRYPTED_PASSTHROUGH_BAD_USER_RESP = {
        "error_code": SmartErrorCode.SESSION_EXPIRED.value,
        "result": {
            "data": {
                "code": SmartErrorCode.BAD_USERNAME.value,
                "encrypt_type": ["1", "2"],
                "key": "Someb64keyWithUnknownPurpose",
                "nonce": "MixedCaseAlphaNumericWithUnknownPurpose",
            }
        },
    }

    UNENCRYPTED_PASSTHROUGH_HANDSHAKE_RESP = {
        "error_code": SmartErrorCode.SESSION_EXPIRED.value,
        "result": {
            "data": {
                "code": SmartErrorCode.SESSION_EXPIRED.value,
                "time": 9,
                "max_time": 10,
                "sec_left": 0,
                "encrypt_type": ["1", "2"],
                "key": "Someb64keyWithUnknownPurpose",
                "nonce": "MixedCaseAlphaNumericWithUnknownPurpose",
            }
        },
    }

    UNENCRYPTED_PASSTHROUGH_GOOD_LOGIN_RESPONSE = {
        "error_code": 0,
        "result": {"stok": MOCK_UNENCRYPTED_PASSTHROUGH_STOK, "user_group": "root"},
    }

    class _mock_response:
        def __init__(self, status, request: dict):
            self.status = status
            self._json = request

        async def __aenter__(self):
            return self

        async def __aexit__(self, exc_t, exc_v, exc_tb):
            pass

        async def read(self):
            if isinstance(self._json, dict):
                return json_dumps(self._json).encode()
            return self._json

    def __init__(
        self,
        host,
        *,
        status_code=200,
        status_code_list=None,
        want_default_username: bool = False,
        do_not_encrypt_response=False,
        send_response=None,
        sequential_request_delay=0,
        send_error_code=0,
        secure_passthrough_error_code=0,
        digest_password_fail=False,
        device_blocked=False,
        unencrypted_passthrough=False,
    ):
        self.host = host
        self.http_client = HttpClient(DeviceConfig(self.host))
        self.encryption_session: AesEncyptionSession | None = None
        self.server_nonce = secrets.token_bytes(8).hex().upper()
        self.handshake1_complete = False

        # test behaviour attributes
        self.status_code = status_code
        self.status_code_list = status_code_list if status_code_list else []
        self.send_error_code = send_error_code
        self.secure_passthrough_error_code = secure_passthrough_error_code
        self.do_not_encrypt_response = do_not_encrypt_response
        self.want_default_username = want_default_username
        self.digest_password_fail = digest_password_fail
        self.device_blocked = device_blocked
        self.unencrypted_passthrough = unencrypted_passthrough

        self._next_responses: list[dict | bytes] = []

    def _get_status_code(self):
        if self.status_code_list:
            return self.status_code_list.pop(0)
        return self.status_code

    async def post(self, url: URL, params=None, json=None, data=None, *_, **__):
        if data:
            json = json_loads(data)
        res = await self._post(url, json)
        return res

    async def _post(self, url: URL, json: dict[str, Any]):
        method = json["method"]

        if method == "login" and not self.handshake1_complete:
            return await self._return_handshake1_response(url, json)

        if method == "login" and self.handshake1_complete:
            if self.unencrypted_passthrough:
                return await self._return_unencrypted_passthrough_login_response(
                    url, json
                )

            return await self._return_handshake2_response(url, json)
        elif method == "securePassthrough":
            assert url == URL(f"https://{self.host}/stok={MOCK_STOCK}/ds")
            return await self._return_secure_passthrough_response(url, json)
        else:
            # The unencrypted passthrough with have actual query method names.
            # This path is also used by the mock class to return unencrypted
            # responses to single 'get' queries which the secure fw returns as unencrypted
            stok = (
                MOCK_UNENCRYPTED_PASSTHROUGH_STOK
                if self.unencrypted_passthrough
                else MOCK_STOCK
            )
            assert url == URL(f"https://{self.host}/stok={stok}/ds")
            return await self._return_send_response(url, json)

    async def _return_handshake1_response(self, url: URL, request: dict[str, Any]):
        request_nonce = request["params"].get("cnonce")
        request_username = request["params"].get("username")

        if self.device_blocked:
            return self._mock_response(self.status_code, self.DEVICE_BLOCKED_RESP)

        if (self.want_default_username and request_username != MOCK_ADMIN_USER) or (
            not self.want_default_username and request_username != MOCK_USER
        ):
            resp = (
                self.UNENCRYPTED_PASSTHROUGH_BAD_USER_RESP
                if self.unencrypted_passthrough
                else self.BAD_USER_RESP
            )
            return self._mock_response(self.status_code, resp)

        device_confirm = SslAesTransport.generate_confirm_hash(
            request_nonce, self.server_nonce, _sha256_hash(MOCK_PWD.encode())
        )
        self.handshake1_complete = True

        if self.unencrypted_passthrough:
            return self._mock_response(
                self._get_status_code(), self.UNENCRYPTED_PASSTHROUGH_HANDSHAKE_RESP
            )

        resp = {
            "error_code": SmartErrorCode.INVALID_NONCE.value,
            "result": {
                "data": {
                    "code": SmartErrorCode.INVALID_NONCE.value,
                    "encrypt_type": ["3"],
                    "key": "Someb64keyWithUnknownPurpose",
                    "nonce": self.server_nonce,
                    "device_confirm": device_confirm,
                }
            },
        }
        return self._mock_response(self._get_status_code(), resp)

    async def _return_unencrypted_passthrough_login_response(
        self, url: URL, request: dict[str, Any]
    ):
        request_username = request["params"].get("username")
        request_password = request["params"].get("password")
        if (self.want_default_username and request_username != MOCK_ADMIN_USER) or (
            not self.want_default_username and request_username != MOCK_USER
        ):
            return self._mock_response(
                self._get_status_code(), self.UNENCRYPTED_PASSTHROUGH_BAD_USER_RESP
            )

        expected_pwd = _md5_hash(MOCK_PWD.encode())
        if request_password != expected_pwd or self.digest_password_fail:
            return self._mock_response(
                self._get_status_code(), self.UNENCRYPTED_PASSTHROUGH_HANDSHAKE_RESP
            )

        return self._mock_response(
            self._get_status_code(), self.UNENCRYPTED_PASSTHROUGH_GOOD_LOGIN_RESPONSE
        )

    async def _return_handshake2_response(self, url: URL, request: dict[str, Any]):
        request_nonce = request["params"].get("cnonce")
        request_username = request["params"].get("username")
        if (self.want_default_username and request_username != MOCK_ADMIN_USER) or (
            not self.want_default_username and request_username != MOCK_USER
        ):
            return self._mock_response(self._get_status_code(), self.BAD_USER_RESP)

        request_password = request["params"].get("digest_passwd")
        expected_pwd = SslAesTransport.generate_digest_password(
            request_nonce, self.server_nonce, _sha256_hash(MOCK_PWD.encode())
        )
        if request_password != expected_pwd or self.digest_password_fail:
            return self._mock_response(self._get_status_code(), self.BAD_PWD_RESP)

        lsk = SslAesTransport.generate_encryption_token(
            "lsk", request_nonce, self.server_nonce, _sha256_hash(MOCK_PWD.encode())
        )
        ivb = SslAesTransport.generate_encryption_token(
            "ivb", request_nonce, self.server_nonce, _sha256_hash(MOCK_PWD.encode())
        )
        self.encryption_session = AesEncyptionSession(lsk, ivb)
        resp = {
            "error_code": 0,
            "result": {"stok": MOCK_STOCK, "user_group": "root", "start_seq": 100},
        }
        return self._mock_response(self._get_status_code(), resp)

    async def _return_secure_passthrough_response(self, url: URL, json: dict[str, Any]):
        encrypted_request = json["params"]["request"]
        assert self.encryption_session
        decrypted_request = self.encryption_session.decrypt(encrypted_request.encode())
        decrypted_request_dict = json_loads(decrypted_request)

        if self._next_responses:
            next_response = self._next_responses.pop(0)
            if isinstance(next_response, dict):
                decrypted_response_data = json_dumps(next_response).encode()
                encrypted_response = self.encryption_session.encrypt(
                    decrypted_response_data
                )
            else:
                encrypted_response = next_response
        else:
            decrypted_response = await self._post(url, decrypted_request_dict)
            async with decrypted_response:
                decrypted_response_data = await decrypted_response.read()
            encrypted_response = self.encryption_session.encrypt(
                decrypted_response_data
            )

        response = (
            decrypted_response_data
            if self.do_not_encrypt_response
            else encrypted_response
        )
        result = {
            "result": {"response": response.decode()},
            "error_code": self.secure_passthrough_error_code,
        }
        return self._mock_response(self._get_status_code(), result)

    async def _return_send_response(self, url: URL, json: dict[str, Any]):
        result = {"result": {"method": None}, "error_code": self.send_error_code}
        return self._mock_response(self._get_status_code(), result)

    def put_next_response(self, request: dict | bytes) -> None:
        self._next_responses.append(request)



"""Implementation of battery module."""

from __future__ import annotations

from typing import Annotated

from ...exceptions import KasaException
from ...feature import Feature
from ...module import FeatureAttribute
from ..smartmodule import SmartModule


class BatterySensor(SmartModule):
    """Implementation of battery module."""

    REQUIRED_COMPONENT = "battery_detect"
    QUERY_GETTER_NAME = "get_battery_detect_info"

    def _initialize_features(self) -> None:
        """Initialize features."""
        if (
            "at_low_battery" in self._device.sys_info
            or "is_low" in self._device.sys_info
        ):
            self._add_feature(
                Feature(
                    self._device,
                    "battery_low",
                    "Battery low",
                    container=self,
                    attribute_getter="battery_low",
                    icon="mdi:alert",
                    type=Feature.Type.BinarySensor,
                    category=Feature.Category.Debug,
                )
            )

        # Some devices, like T110 contact sensor do not report the battery percentage
        if "battery_percentage" in self._device.sys_info:
            self._add_feature(
                Feature(
                    self._device,
                    "battery_level",
                    "Battery level",
                    container=self,
                    attribute_getter="battery",
                    icon="mdi:battery",
                    unit_getter=lambda: "%",
                    category=Feature.Category.Info,
                    type=Feature.Type.Sensor,
                )
            )

    def query(self) -> dict:
        """Query to execute during the update cycle."""
        return {}

    @property
    def battery(self) -> Annotated[int, FeatureAttribute()]:
        """Return battery level."""
        return self._device.sys_info["battery_percentage"]

    @property
    def battery_low(self) -> Annotated[bool, FeatureAttribute()]:
        """Return True if battery is low."""
        is_low = self._device.sys_info.get(
            "at_low_battery", self._device.sys_info.get("is_low")
        )
        if is_low is None:
            raise KasaException("Device does not report battery low status")

        return is_low



"""Tests for smart camera devices."""

from __future__ import annotations

import base64
import json
from unittest.mock import patch

import pytest

from kasa import Credentials, Device, DeviceType, Module, StreamResolution

from ...conftest import device_smartcam, parametrize

not_child_camera_smartcam = parametrize(
    "not child camera smartcam",
    device_type_filter=[DeviceType.Camera],
    protocol_filter={"SMARTCAM"},
)


@device_smartcam
async def test_state(dev: Device):
    if dev.device_type is DeviceType.Hub:
        pytest.skip("Hubs cannot be switched on and off")

    state = dev.is_on
    await dev.set_state(not state)
    await dev.update()
    assert dev.is_on is not state


@not_child_camera_smartcam
async def test_stream_rtsp_url(dev: Device):
    camera_module = dev.modules.get(Module.Camera)
    assert camera_module

    await camera_module.set_state(True)
    await dev.update()
    assert camera_module.is_on
    url = camera_module.stream_rtsp_url(Credentials("foo", "bar"))
    assert url == "rtsp://foo:bar@127.0.0.123:554/stream1"

    url = camera_module.stream_rtsp_url(
        Credentials("foo", "bar"), stream_resolution=StreamResolution.HD
    )
    assert url == "rtsp://foo:bar@127.0.0.123:554/stream1"

    url = camera_module.stream_rtsp_url(
        Credentials("foo", "bar"), stream_resolution=StreamResolution.SD
    )
    assert url == "rtsp://foo:bar@127.0.0.123:554/stream2"

    with patch.object(dev.config, "credentials", Credentials("bar", "foo")):
        url = camera_module.stream_rtsp_url()
    assert url == "rtsp://bar:foo@127.0.0.123:554/stream1"

    with patch.object(dev.config, "credentials", Credentials("bar", "")):
        url = camera_module.stream_rtsp_url()
    assert url is None

    with patch.object(dev.config, "credentials", Credentials("", "Foo")):
        url = camera_module.stream_rtsp_url()
    assert url is None

    # Test with credentials_hash
    cred = json.dumps({"un": "bar", "pwd": "foobar"})
    cred_hash = base64.b64encode(cred.encode()).decode()
    with (
        patch.object(dev.config, "credentials", None),
        patch.object(dev.config, "credentials_hash", cred_hash),
    ):
        url = camera_module.stream_rtsp_url()
    assert url == "rtsp://bar:foobar@127.0.0.123:554/stream1"

    # Test with invalid credentials_hash
    with (
        patch.object(dev.config, "credentials", None),
        patch.object(dev.config, "credentials_hash", b"238472871"),
    ):
        url = camera_module.stream_rtsp_url()
    assert url is None

    # Test with no credentials
    with (
        patch.object(dev.config, "credentials", None),
        patch.object(dev.config, "credentials_hash", None),
    ):
        url = camera_module.stream_rtsp_url()
    assert url is None


@not_child_camera_smartcam
async def test_onvif_url(dev: Device):
    """Test the onvif url."""
    camera_module = dev.modules.get(Module.Camera)
    assert camera_module

    url = camera_module.onvif_url()
    assert url == "http://127.0.0.123:2020/onvif/device_service"



"""Tests for all devices."""

from __future__ import annotations

import importlib
import inspect
import pkgutil
import sys
import zoneinfo
from contextlib import AbstractContextManager, nullcontext
from unittest.mock import AsyncMock, patch

import pytest

import kasa
from kasa import Credentials, Device, DeviceConfig, DeviceType, KasaException, Module
from kasa.iot import (
    IotBulb,
    IotCamera,
    IotDevice,
    IotDimmer,
    IotLightStrip,
    IotPlug,
    IotStrip,
    IotWallSwitch,
)
from kasa.iot.iottimezone import (
    TIMEZONE_INDEX,
    get_timezone,
    get_timezone_index,
)
from kasa.iot.modules import IotLightPreset
from kasa.smart import SmartChildDevice, SmartDevice
from kasa.smartcam import SmartCamChild, SmartCamDevice


def _get_subclasses(of_class):
    package = sys.modules["kasa"]
    subclasses = set()
    for _, modname, _ in pkgutil.iter_modules(package.__path__):
        importlib.import_module("." + modname, package="kasa")
        module = sys.modules["kasa." + modname]
        for name, obj in inspect.getmembers(module):
            if (
                inspect.isclass(obj)
                and issubclass(obj, of_class)
                and module.__package__ != "kasa"
                and module.__package__ != "kasa.interfaces"
            ):
                subclasses.add((module.__package__ + "." + name, obj))
    return sorted(subclasses)


device_classes = pytest.mark.parametrize(
    "device_class_name_obj", _get_subclasses(Device), ids=lambda t: t[0]
)


async def test_device_id(dev: Device):
    """Test all devices have a device id."""
    assert dev.device_id


async def test_alias(dev):
    test_alias = "TEST1234"
    original = dev.alias

    assert isinstance(original, str | None)
    await dev.set_alias(test_alias)
    await dev.update()
    assert dev.alias == test_alias

    # If alias is None set it back to empty string
    await dev.set_alias(original or "")
    await dev.update()
    assert dev.alias == original


@device_classes
async def test_device_class_ctors(device_class_name_obj):
    """Make sure constructor api not broken for new and existing SmartDevices."""
    host = "127.0.0.2"
    port = 1234
    credentials = Credentials("foo", "bar")
    config = DeviceConfig(host, port_override=port, credentials=credentials)
    klass = device_class_name_obj[1]
    if issubclass(klass, SmartChildDevice | SmartCamChild):
        parent = SmartDevice(host, config=config)
        smartcam_required = {
            "device_model": "foo",
            "device_type": "SMART.TAPODOORBELL",
            "alias": "Foo",
            "sw_ver": "1.1",
            "hw_ver": "1.0",
            "mac": "1.2.3.4",
            "hwId": "hw_id",
            "oem_id": "oem_id",
        }
        dev = klass(
            parent,
            {"dummy": "info", "device_id": "dummy", **smartcam_required},
            {
                "component_list": [{"id": "device", "ver_code": 1}],
                "app_component_list": [{"name": "device", "version": 1}],
            },
        )
    else:
        dev = klass(host, config=config)
    assert dev.host == host
    assert dev.port == port
    assert dev.credentials == credentials


@device_classes
async def test_device_class_repr(device_class_name_obj):
    """Test device repr when update() not called and no discovery info."""
    host = "127.0.0.2"
    port = 1234
    credentials = Credentials("foo", "bar")
    config = DeviceConfig(host, port_override=port, credentials=credentials)
    klass = device_class_name_obj[1]
    if issubclass(klass, SmartChildDevice | SmartCamChild):
        parent = SmartDevice(host, config=config)
        dev = klass(
            parent,
            {"dummy": "info", "device_id": "dummy"},
            {
                "component_list": [{"id": "device", "ver_code": 1}],
                "app_component_list": [{"name": "device", "version": 1}],
            },
        )
    else:
        dev = klass(host, config=config)

    CLASS_TO_DEFAULT_TYPE = {
        IotDevice: DeviceType.Unknown,
        IotBulb: DeviceType.Bulb,
        IotPlug: DeviceType.Plug,
        IotDimmer: DeviceType.Dimmer,
        IotStrip: DeviceType.Strip,
        IotWallSwitch: DeviceType.WallSwitch,
        IotLightStrip: DeviceType.LightStrip,
        IotCamera: DeviceType.Camera,
        SmartChildDevice: DeviceType.Unknown,
        SmartDevice: DeviceType.Unknown,
        SmartCamDevice: DeviceType.Unknown,
        SmartCamChild: DeviceType.Unknown,
    }
    type_ = CLASS_TO_DEFAULT_TYPE[klass]
    child_repr = "<DeviceType.Unknown(child) of <DeviceType.Unknown at 127.0.0.2 - update() needed>>"
    not_child_repr = f"<{type_} at 127.0.0.2 - update() needed>"
    expected_repr = (
        child_repr if klass in {SmartChildDevice, SmartCamChild} else not_child_repr
    )
    assert repr(dev) == expected_repr


async def test_create_device_with_timeout():
    """Make sure timeout is passed to the protocol."""
    host = "127.0.0.1"
    dev = IotDevice(host, config=DeviceConfig(host, timeout=100))
    assert dev.protocol._transport._timeout == 100
    dev = SmartDevice(host, config=DeviceConfig(host, timeout=100))
    assert dev.protocol._transport._timeout == 100


async def test_create_thin_wrapper():
    """Make sure thin wrapper is created with the correct device type."""
    mock = AsyncMock()
    config = DeviceConfig(
        host="test_host",
        port_override=1234,
        timeout=100,
        credentials=Credentials("username", "password"),
    )
    with patch("kasa.device_factory.connect", return_value=mock) as connect:
        dev = await Device.connect(config=config)
        assert dev is mock

    connect.assert_called_once_with(
        host=None,
        config=config,
    )


@pytest.mark.parametrize(
    ("device_class", "use_class"), kasa.deprecated_smart_devices.items()
)
def test_deprecated_devices(device_class, use_class):
    package_name = ".".join(use_class.__module__.split(".")[:-1])
    msg = f"{device_class} is deprecated, use {use_class.__name__} from package {package_name} instead"
    with pytest.deprecated_call(match=msg):
        getattr(kasa, device_class)
    packages = package_name.split(".")
    module = __import__(packages[0])
    for _ in packages[1:]:
        module = importlib.import_module(package_name, package=module.__name__)
    getattr(module, use_class.__name__)


@pytest.mark.parametrize(
    ("deprecated_class", "use_class"), kasa.deprecated_classes.items()
)
def test_deprecated_classes(deprecated_class, use_class):
    msg = f"{deprecated_class} is deprecated, use {use_class.__name__} instead"
    with pytest.deprecated_call(match=msg):
        getattr(kasa, deprecated_class)
    getattr(kasa, use_class.__name__)


deprecated_is_device_type = {
    "is_bulb": DeviceType.Bulb,
    "is_plug": DeviceType.Plug,
    "is_dimmer": DeviceType.Dimmer,
    "is_light_strip": DeviceType.LightStrip,
    "is_wallswitch": DeviceType.WallSwitch,
    "is_strip": DeviceType.Strip,
    "is_strip_socket": DeviceType.StripSocket,
}
deprecated_is_light_function_smart_module = {
    "is_color": "Color",
    "is_dimmable": "Brightness",
    "is_variable_color_temp": "ColorTemperature",
}


def test_deprecated_device_type_attributes(dev: SmartDevice):
    """Test deprecated attributes on all devices."""

    def _test_attr(attribute):
        msg = f"{attribute} is deprecated"
        if module := Device._deprecated_device_type_attributes[attribute][0]:
            msg += f", use: {module} in device.modules instead"
        with pytest.deprecated_call(match=msg):
            val = getattr(dev, attribute)
        return val

    for attribute in deprecated_is_device_type:
        val = _test_attr(attribute)
        expected_val = dev.device_type == deprecated_is_device_type[attribute]
        assert val == expected_val


async def _test_attribute(
    dev: Device, attribute_name, is_expected, module_name, *args, will_raise=False
):
    if is_expected and will_raise:
        ctx: AbstractContextManager | nullcontext = pytest.raises(will_raise)
        dep_context: pytest.WarningsRecorder | nullcontext = pytest.deprecated_call(
            match=(f"{attribute_name} is deprecated, use:")
        )
    elif is_expected:
        ctx = nullcontext()
        dep_context = pytest.deprecated_call(
            match=(f"{attribute_name} is deprecated, use:")
        )
    else:
        ctx = pytest.raises(
            AttributeError, match=f"Device has no attribute '{attribute_name}'"
        )
        dep_context = nullcontext()

    with dep_context, ctx:
        if args:
            await getattr(dev, attribute_name)(*args)
        else:
            attribute_val = getattr(dev, attribute_name)
            assert attribute_val is not None


async def test_deprecated_light_effect_attributes(dev: Device):
    light_effect = dev.modules.get(Module.LightEffect)

    await _test_attribute(dev, "effect", bool(light_effect), "LightEffect")
    await _test_attribute(dev, "effect_list", bool(light_effect), "LightEffect")
    await _test_attribute(dev, "set_effect", bool(light_effect), "LightEffect", "Off")
    exc = (
        NotImplementedError
        if light_effect and not light_effect.has_custom_effects
        else None
    )
    await _test_attribute(
        dev,
        "set_custom_effect",
        bool(light_effect),
        "LightEffect",
        {"enable": 0, "name": "foo", "id": "bar"},
        will_raise=exc,
    )


async def test_deprecated_light_attributes(dev: Device):
    light = dev.modules.get(Module.Light)

    await _test_attribute(dev, "is_dimmable", bool(light), "Light")
    await _test_attribute(dev, "is_color", bool(light), "Light")
    await _test_attribute(dev, "is_variable_color_temp", bool(light), "Light")

    exc = KasaException if light and not light.has_feature("brightness") else None
    await _test_attribute(dev, "brightness", bool(light), "Light", will_raise=exc)
    await _test_attribute(
        dev, "set_brightness", bool(light), "Light", 50, will_raise=exc
    )

    exc = KasaException if light and not light.has_feature("hsv") else None
    await _test_attribute(dev, "hsv", bool(light), "Light", will_raise=exc)
    await _test_attribute(
        dev, "set_hsv", bool(light), "Light", 50, 50, 50, will_raise=exc
    )

    exc = KasaException if light and not light.has_feature("color_temp") else None
    await _test_attribute(dev, "color_temp", bool(light), "Light", will_raise=exc)
    await _test_attribute(
        dev, "set_color_temp", bool(light), "Light", 2700, will_raise=exc
    )
    await _test_attribute(
        dev, "valid_temperature_range", bool(light), "Light", will_raise=exc
    )

    await _test_attribute(dev, "has_effects", bool(light), "Light")


async def test_deprecated_other_attributes(dev: Device):
    led_module = dev.modules.get(Module.Led)

    await _test_attribute(dev, "led", bool(led_module), "Led")
    await _test_attribute(dev, "set_led", bool(led_module), "Led", True)
    await _test_attribute(dev, "supported_modules", True, None)


async def test_deprecated_emeter_attributes(dev: Device):
    energy_module = dev.modules.get(Module.Energy)

    await _test_attribute(dev, "get_emeter_realtime", bool(energy_module), "Energy")
    await _test_attribute(dev, "emeter_realtime", bool(energy_module), "Energy")
    await _test_attribute(dev, "emeter_today", bool(energy_module), "Energy")
    await _test_attribute(dev, "emeter_this_month", bool(energy_module), "Energy")
    await _test_attribute(dev, "current_consumption", bool(energy_module), "Energy")
    await _test_attribute(dev, "get_emeter_daily", bool(energy_module), "Energy")
    await _test_attribute(dev, "get_emeter_monthly", bool(energy_module), "Energy")


async def test_deprecated_light_preset_attributes(dev: Device):
    preset = dev.modules.get(Module.LightPreset)

    exc: type[AttributeError] | type[KasaException] | None = (
        AttributeError if not preset else None
    )
    await _test_attribute(dev, "presets", bool(preset), "LightPreset", will_raise=exc)

    exc = None
    is_expected = bool(preset)
    # deprecated save_preset not implemented for smart devices as it's unlikely anyone
    # has an existing reliance on this for the newer devices.
    if isinstance(dev, SmartDevice):
        is_expected = False

    if preset and len(preset.preset_states_list) == 0:
        exc = KasaException

    await _test_attribute(
        dev,
        "save_preset",
        is_expected,
        "LightPreset",
        IotLightPreset(index=0, hue=100, brightness=100, saturation=0, color_temp=0),  # type: ignore[call-arg]
        will_raise=exc,
    )


async def test_device_type_aliases():
    """Test that the device type aliases in Device work."""

    def _mock_connect(config, *args, **kwargs):
        mock = AsyncMock()
        mock.config = config
        return mock

    with patch("kasa.device_factory.connect", side_effect=_mock_connect):
        dev = await Device.connect(
            config=Device.Config(
                host="127.0.0.1",
                credentials=Device.Credentials(username="user", password="foobar"),  # noqa: S106
                connection_type=Device.ConnectionParameters(
                    device_family=Device.Family.SmartKasaPlug,
                    encryption_type=Device.EncryptionType.Klap,
                    login_version=2,
                ),
            )
        )
        assert isinstance(dev.config, DeviceConfig)
        assert DeviceType.Dimmer == Device.Type.Dimmer


async def test_device_timezones():
    """Test the timezone data is good."""
    # Check all indexes return a zoneinfo
    for i in range(110):
        tz = await get_timezone(i)
        assert tz
        assert tz != zoneinfo.ZoneInfo("Etc/UTC"), f"{i} is default Etc/UTC"

    # Check an unexpected index returns a UTC default.
    tz = await get_timezone(110)
    assert tz == zoneinfo.ZoneInfo("Etc/UTC")

    # Get an index from a timezone
    for index, zone in TIMEZONE_INDEX.items():
        zone_info = zoneinfo.ZoneInfo(zone)
        found_index = await get_timezone_index(zone_info)
        assert found_index == index

    # Try a timezone not hardcoded finds another match
    index = await get_timezone_index(zoneinfo.ZoneInfo("Asia/Katmandu"))
    assert index == 77

    # Try a timezone not hardcoded no match
    with pytest.raises(zoneinfo.ZoneInfoNotFoundError):
        await get_timezone_index(zoneinfo.ZoneInfo("Foo/bar"))



"""Implementation of matter module."""

from __future__ import annotations

from ...feature import Feature
from ..smartcammodule import SmartCamModule


class Matter(SmartCamModule):
    """Implementation of matter module."""

    QUERY_GETTER_NAME = "getMatterSetupInfo"
    QUERY_MODULE_NAME = "matter"
    REQUIRED_COMPONENT = "matter"

    def _initialize_features(self) -> None:
        """Initialize features after the initial update."""
        self._add_feature(
            Feature(
                self._device,
                id="matter_setup_code",
                name="Matter setup code",
                container=self,
                attribute_getter=lambda x: x.info["setup_code"],
                type=Feature.Type.Sensor,
                category=Feature.Category.Debug,
            )
        )
        self._add_feature(
            Feature(
                self._device,
                id="matter_setup_payload",
                name="Matter setup payload",
                container=self,
                attribute_getter=lambda x: x.info["setup_payload"],
                type=Feature.Type.Sensor,
                category=Feature.Category.Debug,
            )
        )

    @property
    def info(self) -> dict[str, str]:
        """Matter setup info."""
        return self.data



"""Tests for smartcam motion detection module."""

from __future__ import annotations

from kasa import Device
from kasa.smartcam.smartcammodule import SmartCamModule

from ...device_fixtures import parametrize

motiondetection = parametrize(
    "has motion detection", component_filter="detection", protocol_filter={"SMARTCAM"}
)


@motiondetection
async def test_motiondetection(dev: Device):
    """Test device motion detection."""
    motion = dev.modules.get(SmartCamModule.SmartCamMotionDetection)
    assert motion

    mde_feat = dev.features.get("motion_detection")
    assert mde_feat

    original_enabled = motion.enabled

    try:
        await motion.set_enabled(not original_enabled)
        await dev.update()
        assert motion.enabled is not original_enabled
        assert mde_feat.value is not original_enabled

        await motion.set_enabled(original_enabled)
        await dev.update()
        assert motion.enabled is original_enabled
        assert mde_feat.value is original_enabled

        await mde_feat.set_value(not original_enabled)
        await dev.update()
        assert motion.enabled is not original_enabled
        assert mde_feat.value is not original_enabled

    finally:
        await motion.set_enabled(original_enabled)



"""Module for bulbs (LB*, KL*, KB*)."""

from __future__ import annotations

import logging
import re
from dataclasses import dataclass
from enum import Enum
from typing import Annotated, cast

from mashumaro import DataClassDictMixin
from mashumaro.config import BaseConfig
from mashumaro.types import Alias

from ..device_type import DeviceType
from ..deviceconfig import DeviceConfig
from ..interfaces.light import HSV, ColorTempRange
from ..module import Module
from ..protocols import BaseProtocol
from .iotdevice import IotDevice, KasaException, requires_update
from .modules import (
    Antitheft,
    Cloud,
    Countdown,
    Emeter,
    Light,
    LightPreset,
    Schedule,
    Time,
    Usage,
)


class BehaviorMode(str, Enum):
    """Enum to present type of turn on behavior."""

    #: Return to the last state known state.
    Last = "last_status"
    #: Use chosen preset.
    Preset = "customize_preset"
    #: Circadian
    Circadian = "circadian"


@dataclass
class TurnOnBehavior(DataClassDictMixin):
    """Model to present a single turn on behavior.

    :param int preset: the index number of wanted preset.
    :param BehaviorMode mode: last status or preset mode.
     If you are changing existing settings, you should not set this manually.

    To change the behavior, it is only necessary to change the :attr:`preset` field
    to contain either the preset index, or ``None`` for the last known state.
    """

    class Config(BaseConfig):
        """Serialization config."""

        omit_none = True
        serialize_by_alias = True

    #: Wanted behavior
    mode: BehaviorMode
    #: Index of preset to use, or ``None`` for the last known state.
    preset: Annotated[int | None, Alias("index")] = None
    brightness: int | None = None
    color_temp: int | None = None
    hue: int | None = None
    saturation: int | None = None


@dataclass
class TurnOnBehaviors(DataClassDictMixin):
    """Model to contain turn on behaviors."""

    #: The behavior when the bulb is turned on programmatically.
    soft: Annotated[TurnOnBehavior, Alias("soft_on")]
    #: The behavior when the bulb has been off from mains power.
    hard: Annotated[TurnOnBehavior, Alias("hard_on")]


TPLINK_KELVIN = {
    "LB130": ColorTempRange(2500, 9000),
    "LB120": ColorTempRange(2700, 6500),
    "LB230": ColorTempRange(2500, 9000),
    "KB130": ColorTempRange(2500, 9000),
    "KL130": ColorTempRange(2500, 9000),
    "KL125": ColorTempRange(2500, 6500),
    "KL135": ColorTempRange(2500, 9000),
    r"KL120\(EU\)": ColorTempRange(2700, 6500),
    r"KL120\(US\)": ColorTempRange(2700, 5000),
    r"KL430": ColorTempRange(2500, 9000),
}


NON_COLOR_MODE_FLAGS = {"transition_period", "on_off"}

_LOGGER = logging.getLogger(__name__)


class IotBulb(IotDevice):
    r"""Representation of a TP-Link Smart Bulb.

    To initialize, you have to await :func:`update()` at least once.
    This will allow accessing the properties using the exposed properties.

    All changes to the device are done using awaitable methods,
    which will not change the cached values,
    so you must await :func:`update()` to fetch updates values from the device.

    Errors reported by the device are raised as
    :class:`KasaException <kasa.exceptions.KasaException>`,
    and should be handled by the user of the library.

    Examples:
        >>> import asyncio
        >>> bulb = IotBulb("127.0.0.1")
        >>> asyncio.run(bulb.update())
        >>> print(bulb.alias)
        Bulb2

        Bulbs, like any other supported devices, can be turned on and off:

        >>> asyncio.run(bulb.turn_off())
        >>> asyncio.run(bulb.turn_on())
        >>> asyncio.run(bulb.update())
        >>> print(bulb.is_on)
        True

        You can use the ``is_``-prefixed properties to check for supported features:

        >>> bulb.is_dimmable
        True
        >>> bulb.is_color
        True
        >>> bulb.is_variable_color_temp
        True

        All known bulbs support changing the brightness:

        >>> bulb.brightness
        30
        >>> asyncio.run(bulb.set_brightness(50))
        >>> asyncio.run(bulb.update())
        >>> bulb.brightness
        50

        Bulbs supporting color temperature can be queried for the supported range:

        >>> bulb.valid_temperature_range
        ColorTempRange(min=2500, max=9000)
        >>> asyncio.run(bulb.set_color_temp(3000))
        >>> asyncio.run(bulb.update())
        >>> bulb.color_temp
        3000

        Color bulbs can be adjusted by passing hue, saturation and value:

        >>> asyncio.run(bulb.set_hsv(180, 100, 80))
        >>> asyncio.run(bulb.update())
        >>> bulb.hsv
        HSV(hue=180, saturation=100, value=80)

        If you don't want to use the default transitions,
        you can pass `transition` in milliseconds.
        All methods changing the state of the device support this parameter:

        * :func:`turn_on`
        * :func:`turn_off`
        * :func:`set_hsv`
        * :func:`set_color_temp`
        * :func:`set_brightness`

        Light strips (e.g., KL420L5) do not support this feature,
        but silently ignore the parameter.
        The following changes the brightness over a period of 10 seconds:

        >>> asyncio.run(bulb.set_brightness(100, transition=10_000))

        Bulb configuration presets can be accessed using the :func:`presets` property:

        >>> [ preset.to_dict() for preset in bulb.presets }
        [{'brightness': 50, 'hue': 0, 'saturation': 0, 'color_temp': 2700, 'index': 0}, {'brightness': 100, 'hue': 0, 'saturation': 75, 'color_temp': 0, 'index': 1}, {'brightness': 100, 'hue': 120, 'saturation': 75, 'color_temp': 0, 'index': 2}, {'brightness': 100, 'hue': 240, 'saturation': 75, 'color_temp': 0, 'index': 3}]

        To modify an existing preset, pass :class:`~kasa.interfaces.light.LightPreset`
        instance to :func:`save_preset` method:

        >>> preset = bulb.presets[0]
        >>> preset.brightness
        50
        >>> preset.brightness = 100
        >>> asyncio.run(bulb.save_preset(preset))
        >>> asyncio.run(bulb.update())
        >>> bulb.presets[0].brightness
        100

    """  # noqa: E501

    LIGHT_SERVICE = "smartlife.iot.smartbulb.lightingservice"
    SET_LIGHT_METHOD = "transition_light_state"
    emeter_type = "smartlife.iot.common.emeter"

    def __init__(
        self,
        host: str,
        *,
        config: DeviceConfig | None = None,
        protocol: BaseProtocol | None = None,
    ) -> None:
        super().__init__(host=host, config=config, protocol=protocol)
        self._device_type = DeviceType.Bulb

    async def _initialize_modules(self) -> None:
        """Initialize modules not added in init."""
        await super()._initialize_modules()
        self.add_module(
            Module.IotSchedule, Schedule(self, "smartlife.iot.common.schedule")
        )
        self.add_module(Module.IotUsage, Usage(self, "smartlife.iot.common.schedule"))
        self.add_module(
            Module.IotAntitheft, Antitheft(self, "smartlife.iot.common.anti_theft")
        )
        self.add_module(Module.Time, Time(self, "smartlife.iot.common.timesetting"))
        self.add_module(Module.Energy, Emeter(self, self.emeter_type))
        self.add_module(Module.IotCountdown, Countdown(self, "countdown"))
        self.add_module(Module.IotCloud, Cloud(self, "smartlife.iot.common.cloud"))
        self.add_module(Module.Light, Light(self, self.LIGHT_SERVICE))
        self.add_module(Module.LightPreset, LightPreset(self, self.LIGHT_SERVICE))

    @property  # type: ignore
    @requires_update
    def _is_color(self) -> bool:
        """Whether the bulb supports color changes."""
        sys_info = self.sys_info
        return bool(sys_info["is_color"])

    @property  # type: ignore
    @requires_update
    def _is_dimmable(self) -> bool:
        """Whether the bulb supports brightness changes."""
        sys_info = self.sys_info
        return bool(sys_info["is_dimmable"])

    @property  # type: ignore
    @requires_update
    def _is_variable_color_temp(self) -> bool:
        """Whether the bulb supports color temperature changes."""
        sys_info = self.sys_info
        return bool(sys_info["is_variable_color_temp"])

    @property  # type: ignore
    @requires_update
    def _valid_temperature_range(self) -> ColorTempRange:
        """Return the device-specific white temperature range (in Kelvin).

        :return: White temperature range in Kelvin (minimum, maximum)
        """
        if not self._is_variable_color_temp:
            raise KasaException("Color temperature not supported")

        for model, temp_range in TPLINK_KELVIN.items():
            sys_info = self.sys_info
            if re.match(model, sys_info["model"]):
                return temp_range

        _LOGGER.warning("Unknown color temperature range, fallback to 2700-5000")
        return ColorTempRange(2700, 5000)

    @property  # type: ignore
    @requires_update
    def light_state(self) -> dict[str, str]:
        """Query the light state."""
        light_state = self.sys_info["light_state"]
        if light_state is None:
            raise KasaException(
                "The device has no light_state or you have not called update()"
            )

        # if the bulb is off, its state is stored under a different key
        # as is_on property depends on on_off itself, we check it here manually
        is_on = light_state["on_off"]
        if not is_on:
            off_state = {**light_state["dft_on_state"], "on_off": is_on}
            return cast(dict, off_state)

        return light_state

    @property  # type: ignore
    @requires_update
    def _has_effects(self) -> bool:
        """Return True if the device supports effects."""
        return "lighting_effect_state" in self.sys_info

    async def get_light_details(self) -> dict[str, int]:
        """Return light details.

        Example::

            {'lamp_beam_angle': 290, 'min_voltage': 220, 'max_voltage': 240,
             'wattage': 5, 'incandescent_equivalent': 40, 'max_lumens': 450,
              'color_rendering_index': 80}
        """
        return await self._query_helper(self.LIGHT_SERVICE, "get_light_details")

    async def get_turn_on_behavior(self) -> TurnOnBehaviors:
        """Return the behavior for turning the bulb on."""
        return TurnOnBehaviors.from_dict(
            await self._query_helper(self.LIGHT_SERVICE, "get_default_behavior")
        )

    async def set_turn_on_behavior(self, behavior: TurnOnBehaviors) -> dict:
        """Set the behavior for turning the bulb on.

        If you do not want to manually construct the behavior object,
        you should use :func:`get_turn_on_behavior` to get the current settings.
        """
        return await self._query_helper(
            self.LIGHT_SERVICE, "set_default_behavior", behavior.to_dict()
        )

    async def get_light_state(self) -> dict[str, dict]:
        """Query the light state."""
        # TODO: add warning and refer to use light.state?
        return await self._query_helper(self.LIGHT_SERVICE, "get_light_state")

    async def _set_light_state(
        self, state: dict, *, transition: int | None = None
    ) -> dict:
        """Set the light state."""
        state = {**state}
        if transition is not None:
            state["transition_period"] = transition

        if "brightness" in state:
            self._raise_for_invalid_brightness(state["brightness"])

        # if no on/off is defined, turn on the light
        if "on_off" not in state:
            state["on_off"] = 1

        # If we are turning on without any color mode flags,
        # we do not want to set ignore_default to ensure
        # we restore the previous state.
        if state["on_off"] and NON_COLOR_MODE_FLAGS.issuperset(state):
            state["ignore_default"] = 0
        else:
            # This is necessary to allow turning on into a specific state
            state["ignore_default"] = 1

        light_state = await self._query_helper(
            self.LIGHT_SERVICE, self.SET_LIGHT_METHOD, state
        )
        return light_state

    @property  # type: ignore
    @requires_update
    def _hsv(self) -> HSV:
        """Return the current HSV state of the bulb.

        :return: hue, saturation and value (degrees, %, %)
        """
        if not self._is_color:
            raise KasaException("Bulb does not support color.")

        light_state = cast(dict, self.light_state)

        hue = light_state["hue"]
        saturation = light_state["saturation"]
        value = self._brightness

        # Simple HSV(hue, saturation, value) is less efficent than below
        # due to the cpython implementation.
        return tuple.__new__(HSV, (hue, saturation, value))

    @requires_update
    async def _set_hsv(
        self,
        hue: int,
        saturation: int,
        value: int | None = None,
        *,
        transition: int | None = None,
    ) -> dict:
        """Set new HSV.

        :param int hue: hue in degrees
        :param int saturation: saturation in percentage [0,100]
        :param int value: value in percentage [0, 100]
        :param int transition: transition in milliseconds.
        """
        if not self._is_color:
            raise KasaException("Bulb does not support color.")

        if not isinstance(hue, int):
            raise TypeError("Hue must be an integer.")
        if not (0 <= hue <= 360):
            raise ValueError(f"Invalid hue value: {hue} (valid range: 0-360)")

        if not isinstance(saturation, int):
            raise TypeError("Saturation must be an integer.")
        if not (0 <= saturation <= 100):
            raise ValueError(
                f"Invalid saturation value: {saturation} (valid range: 0-100%)"
            )

        light_state = {
            "hue": hue,
            "saturation": saturation,
            "color_temp": 0,
        }

        if value is not None:
            self._raise_for_invalid_brightness(value)
            light_state["brightness"] = value

        return await self._set_light_state(light_state, transition=transition)

    @property  # type: ignore
    @requires_update
    def _color_temp(self) -> int:
        """Return color temperature of the device in kelvin."""
        if not self._is_variable_color_temp:
            raise KasaException("Bulb does not support colortemp.")

        light_state = self.light_state
        return int(light_state["color_temp"])

    @requires_update
    async def _set_color_temp(
        self, temp: int, *, brightness: int | None = None, transition: int | None = None
    ) -> dict:
        """Set the color temperature of the device in kelvin.

        :param int temp: The new color temperature, in Kelvin
        :param int transition: transition in milliseconds.
        """
        if not self._is_variable_color_temp:
            raise KasaException("Bulb does not support colortemp.")

        valid_temperature_range = self._valid_temperature_range
        if temp < valid_temperature_range[0] or temp > valid_temperature_range[1]:
            raise ValueError(
                "Temperature should be between {} and {}, was {}".format(
                    *valid_temperature_range, temp
                )
            )

        light_state = {"color_temp": temp}
        if brightness is not None:
            light_state["brightness"] = brightness

        return await self._set_light_state(light_state, transition=transition)

    def _raise_for_invalid_brightness(self, value: int) -> None:
        if not isinstance(value, int):
            raise TypeError("Brightness must be an integer")
        if not (0 <= value <= 100):
            raise ValueError(f"Invalid brightness value: {value} (valid range: 0-100%)")

    @property  # type: ignore
    @requires_update
    def _brightness(self) -> int:
        """Return the current brightness in percentage."""
        if not self._is_dimmable:  # pragma: no cover
            raise KasaException("Bulb is not dimmable.")

        # If the device supports effects and one is active, we get the brightness
        # from the effect. This is not required when setting the brightness as
        # the device handles it via set_light_state
        if (
            light_effect := self.modules.get(Module.IotLightEffect)
        ) is not None and light_effect.effect != light_effect.LIGHT_EFFECTS_OFF:
            return light_effect.brightness
        light_state = self.light_state
        return int(light_state["brightness"])

    @requires_update
    async def _set_brightness(
        self, brightness: int, *, transition: int | None = None
    ) -> dict:
        """Set the brightness in percentage.

        :param int brightness: brightness in percent
        :param int transition: transition in milliseconds.
        """
        if not self._is_dimmable:  # pragma: no cover
            raise KasaException("Bulb is not dimmable.")

        self._raise_for_invalid_brightness(brightness)

        light_state = {"brightness": brightness}
        return await self._set_light_state(light_state, transition=transition)

    @property  # type: ignore
    @requires_update
    def is_on(self) -> bool:
        """Return whether the device is on."""
        light_state = self.light_state
        return bool(light_state["on_off"])

    async def turn_off(self, *, transition: int | None = None, **kwargs) -> dict:
        """Turn the bulb off.

        :param int transition: transition in milliseconds.
        """
        return await self._set_light_state({"on_off": 0}, transition=transition)

    async def turn_on(self, *, transition: int | None = None, **kwargs) -> dict:
        """Turn the bulb on.

        :param int transition: transition in milliseconds.
        """
        return await self._set_light_state({"on_off": 1}, transition=transition)

    @property  # type: ignore
    @requires_update
    def has_emeter(self) -> bool:
        """Return that the bulb has an emeter."""
        return True

    async def set_alias(self, alias: str) -> dict:
        """Set the device name (alias).

        Overridden to use a different module name.
        """
        return await self._query_helper(
            "smartlife.iot.common.system", "set_dev_alias", {"alias": alias}
        )

    @property
    def max_device_response_size(self) -> int:
        """Returns the maximum response size the device can safely construct."""
        return 4096
